#!/bin/env python

#cython: boundscheck=False
#cython: nonecheck=False
#cython: cdivision=True

cdef extern from "math.h":
    double M_PI
    double sqrt(double arg) nogil
    double fabs(double arg) nogil
    double atan2(double y, double x) nogil
    double atan(double arg) nogil
    double sin(double arg) nogil
    double cos(double arg) nogil
    #double INFINITY
    
cdef extern from "float.h":
    double DBL_MAX

cdef:
    INF=(DBL_MAX+DBL_MAX)
    double root2 = sqrt(2.0)
    public unsigned int REFL_RAY=1<<0
    public unsigned int GAUSSLET=1<<1
    public unsigned int PARABASAL=1<<2
    
from libc.stdlib cimport malloc, free, realloc

cdef extern from "stdlib.h" nogil:
    void *memcpy(void *str1, void *str2, size_t n)

import time
import numpy as np
cimport numpy as np_

cdef:
    int NPARA = 6
    

ray_dtype = np.dtype([('origin', np.double, (3,)),
                        ('direction', np.double, (3,)),
                        ('normal', np.double, (3,)),
                        ('E_vector', np.double, (3,)),
                        ('refractive_index', np.complex128),
                        ('E1_amp', np.complex128),
                        ('E2_amp', np.complex128),
                        ('length', np.double),
                        ('phase', np.double),
                        ('accumulated_path', np.double),
                        ('wavelength_idx', np.uint32),
                        ('parent_idx', np.uint32),
                        ('end_face_idx', np.uint32),
                        ('ray_type_id', np.uint32) #No point using a smaller type here as it'll probably get padded by the compiler
                        ])
                        
    
GAUSSLET_ = GAUSSLET
PARABASAL_ = PARABASAL

para_dtype = np.dtype([('origin', np.double, (3,)),
                        ('direction', np.double, (3,)),
                        ('normal', np.double, (3,)),
                        ('length', np.double),
                        ])


gausslet_dtype = np.dtype([
                    ('base_ray', ray_dtype),
                    ('para_rays', para_dtype, (NPARA,))
                    ])


##############################
### Vector maths functions ###
##############################

cdef inline vector_t transform_c(transform_t t, vector_t p) nogil:
    cdef vector_t out
    out.x = p.x*t.m00 + p.y*t.m01 + p.z*t.m02 + t.tx
    out.y = p.x*t.m10 + p.y*t.m11 + p.z*t.m12 + t.ty
    out.z = p.x*t.m20 + p.y*t.m21 + p.z*t.m22 + t.tz
    return out

cdef inline vector_t rotate_c(transform_t t, vector_t p) nogil:
    cdef vector_t out
    out.x = p.x*t.m00 + p.y*t.m01 + p.z*t.m02
    out.y = p.x*t.m10 + p.y*t.m11 + p.z*t.m12
    out.z = p.x*t.m20 + p.y*t.m21 + p.z*t.m22
    return out

cdef inline vector_t set_v(object O):
    cdef vector_t v
    v.x = O[0]
    v.y = O[1]
    v.z = O[2]
    return v

def py_set_v(O):
    cdef vector_t v_
    v_ = set_v(O)
    return (v_.x, v_.y, v_.z)

cdef inline double sep_(vector_t p1, vector_t p2) nogil:
    cdef double a,b
    a = (p2.x-p1.x)
    b = (p2.y-p1.y)
    c = (p2.z-p1.z)
    return sqrt((a*a) + (b*b) + (c*c))

def sep(a, b):
    cdef vector_t a_ = set_v(a), b_ = set_v(b)
    return sep_(a_, b_)

cdef inline vector_t invert_(vector_t v) nogil:
    v.x = -v.x
    v.y = -v.y
    v.z = -v.z
    return v

def invert(v):
    cdef vector_t v_ = set_v(v)
    v_ = invert_(v_)
    return (v_.x, v_.y, v_.z)

cdef inline vector_t multvv_(vector_t a, vector_t b) nogil:
    cdef vector_t out
    out.x = a.x*b.x
    out.y = a.y*b.y
    out.z = a.z*b.z
    return out

def multvv(a, b):
    cdef vector_t a_, b_, c_
    a_ = set_v(a)
    b_ = set_v(b)
    c_ = multvv_(a_, b_)
    return (c_.x, c_.y, c_.z)

cdef inline vector_t multvs_(vector_t a, double b) nogil:
    cdef vector_t out
    out.x = a.x*b
    out.y = a.y*b
    out.z = a.z*b
    return out

def multvs(a, b):
    cdef vector_t a_, c_
    a_ = set_v(a)
    c_ = multvs_(a_, b)
    return (c_.x, c_.y, c_.z)

cdef inline vector_t addvv_(vector_t a, vector_t b) nogil:
    cdef vector_t out
    out.x = a.x+b.x
    out.y = a.y+b.y
    out.z = a.z+b.z
    return out

def addvv(a, b):
    cdef vector_t a_, b_, c_
    a_ = set_v(a)
    b_ = set_v(b)
    c_ = addvv_(a_, b_)
    return (c_.x, c_.y, c_.z)
    
cdef inline vector_t addvs_(vector_t a, double b) nogil:
    cdef vector_t out
    out.x = a.x+b
    out.y = a.y+b
    out.z = a.z+b
    return out

def addvs(a, b):
    cdef vector_t a_, c_
    a_ = set_v(a)
    c_ = addvs_(a_, b)
    return (c_.x, c_.y, c_.z)

cdef inline vector_t subvv_(vector_t a, vector_t b) nogil:
    cdef vector_t out
    out.x = a.x-b.x
    out.y = a.y-b.y
    out.z = a.z-b.z
    return out

def subvv(a, b):
    cdef vector_t a_, b_, c_
    a_ = set_v(a)
    b_ = set_v(b)
    c_ = subvv_(a_, b_)
    return (c_.x, c_.y, c_.z)

cdef inline vector_t subvs_(vector_t a, double b) nogil:
    cdef vector_t out
    out.x = a.x-b
    out.y = a.y-b
    out.z = a.z-b
    return out

def subvs(a, b):
    cdef vector_t a_, c_
    a_ = set_v(a)
    c_ = subvs_(a_, b)
    return (c_.x, c_.y, c_.z)

cdef inline double mag_(vector_t a) nogil:
    return sqrt(a.x*a.x + a.y*a.y + a.z*a.z)

def mag(a):
    cdef vector_t a_
    a_ = set_v(a)
    return mag_(a_)

cdef inline double mag_sq_(vector_t a) nogil:
    return a.x*a.x + a.y*a.y + a.z*a.z

def mag_sq(a):
    cdef vector_t a_
    a_ = set_v(a)
    return mag_sq_(a_)

cdef inline double dotprod_(vector_t a, vector_t b) nogil:
    return a.x*b.x + a.y*b.y + a.z*b.z

def dotprod(a, b):
    cdef vector_t a_, b_
    a_ = set_v(a)
    b_ = set_v(b)
    return dotprod_(a_,b_)

cdef inline vector_t cross_(vector_t a, vector_t b) nogil:
    cdef vector_t c
    c.x = a.y*b.z - a.z*b.y
    c.y = a.z*b.x - a.x*b.z
    c.z = a.x*b.y - a.y*b.x
    return c

def cross(a, b):
    cdef vector_t a_, b_, c_
    a_ = set_v(a)
    b_ = set_v(b)
    c_ = cross_(a_, b_)
    return (c_.x, c_.y, c_.z)

cdef vector_t norm_(vector_t a) nogil:
    cdef double mag=sqrt(a.x*a.x + a.y*a.y + a.z*a.z)
    a.x /= mag
    a.y /= mag
    a.z /= mag
    return a

def norm(a):
    cdef vector_t a_
    a_ = set_v(a)
    a_ = norm_(a_)
    return (a_.x, a_.y, a_.z)

##################################
### Python extension types
##################################

cdef class Transform:
    
    def __init__(self, rotation=[[1,0,0],[0,1,0],[0,0,1]], 
                        translation=[0,0,0]):
        self.rotation = rotation
        self.translation = translation
        
    property rotation:
        def __set__(self, rot):
            cdef transform_t t
            t.m00, t.m01, t.m02 = rot[0]
            t.m10, t.m11, t.m12 = rot[1]
            t.m20, t.m21, t.m22 = rot[2]
            self.trans = t
            
        def __get__(self):
            cdef transform_t t
            t = self.trans
            return [[t.m00, t.m01, t.m02],
                    [t.m10, t.m11, t.m12],
                    [t.m20, t.m21, t.m22]]
    
    property translation:
        def __set__(self, dt):
            self.trans.tx, self.trans.ty, self.trans.tz = dt
        
        def __get__(self):
            return (self.trans.tx, self.trans.ty, self.trans.tz)


cdef class RayCollectionIterator:        
    def __cinit__(self, RayCollection rays):
        self.rays = rays
        self.counter = 0
        
    def __iter__(self):
        return self
    
    def __next__(self):
        cdef Ray ray=Ray.__new__(Ray)
        if self.counter >= self.rays.n_rays:
            raise StopIteration
        ray.ray = self.rays.rays[self.counter]
        self.counter += 1
        return ray
    
    
cdef class GaussletCollectionIterator:        
    def __cinit__(self, GaussletCollection rays):
        self.rays = rays
        self.counter = 0
        
    def __iter__(self):
        return self
    
    def __next__(self):
        cdef Gausslet ray=Gausslet.__new__(Gausslet)
        if self.counter >= self.rays.n_rays:
            raise StopIteration
        ray.gausslet = self.rays.rays[self.counter]
        self.counter += 1
        return ray


cdef class ParabasalRay:
    def __cinit__(self, **kwds):
        self.max_length = 1000.0
        for k in kwds:
            setattr(self, k, kwds[k])
            
    def __repr__(self):
        return "Parabasal Ray(o=%s, d=%s)"%(str(self.origin),
                                            str(self.direction))
        
    property origin:
        """Origin coordinates of the ray"""
        def __get__(self):
            return (self.ray.origin.x,self.ray.origin.y,self.ray.origin.z)
        
        def __set__(self, v):
            self.ray.origin.x = v[0]
            self.ray.origin.y = v[1]
            self.ray.origin.z = v[2]
            
    property direction:
        """direction of the ray, normalised to a unit vector"""
        def __get__(self):
            return (self.ray.direction.x,self.ray.direction.y,self.ray.direction.z)
        
        def __set__(self, v):
            self.ray.direction.x = v[0]
            self.ray.direction.y = v[1]
            self.ray.direction.z = v[2]
            
    property normal:
        """normal vector for the face which created this ray"""
        def __get__(self):
            return (self.ray.normal.x,self.ray.normal.y,self.ray.normal.z)
        
        def __set__(self, v):
            self.ray.normal.x = v[0]
            self.ray.normal.y = v[1]
            self.ray.normal.z = v[2]
            
    property length:
        """The length of the ray. This is infinite in 
        unterminated rays"""
        def __get__(self):
            return self.ray.length
        
        def __set__(self, double v):
            self.ray.length = v
    
    property termination:
        """the end-point of the ray (read only)
        """
        def __get__(self):
            cdef vector_t end
            cdef float length
            if self.ray.length > self.max_length:
                length = self.max_length
            else:
                length = self.ray.length
            end = addvv_(self.ray.origin, multvs_(self.ray.direction, 
                                    length))
            return (end.x, end.y, end.z)


cdef class Ray:
    """ Ray - a wrapper around the ray_t C-structure.
    
    The Ray extension class exists mainly as a convenience for manipulation of single or small numbers of rays 
    from python. Large numbers of rays are more efficiently handled as either RayCollection objects, created in the
    tracing process, or as numpy arrays with the 'ray_dtype' dtype.
    """
    
    def __cinit__(self, **kwds):
        self.max_length = 1000.0
        for k in kwds:
            setattr(self, k, kwds[k])
            
    def __repr__(self):
        return "Ray(o=%s, d=%s)"%(str(self.origin),
                                            str(self.direction))
                
    property origin:
        """Origin coordinates of the ray"""
        def __get__(self):
            return (self.ray.origin.x,self.ray.origin.y,self.ray.origin.z)
        
        def __set__(self, v):
            self.ray.origin.x = v[0]
            self.ray.origin.y = v[1]
            self.ray.origin.z = v[2]
            
    property direction:
        """direction of the ray, normalised to a unit vector"""
        def __get__(self):
            return (self.ray.direction.x,self.ray.direction.y,self.ray.direction.z)
        
        def __set__(self, v):
            self.ray.direction.x = v[0]
            self.ray.direction.y = v[1]
            self.ray.direction.z = v[2]
            
    property normal:
        """normal vector for the face which created this ray"""
        def __get__(self):
            return (self.ray.normal.x,self.ray.normal.y,self.ray.normal.z)
        
        def __set__(self, v):
            self.ray.normal.x = v[0]
            self.ray.normal.y = v[1]
            self.ray.normal.z = v[2]
            
    property E_vector:
        """Unit vector, perpendicular to the ray direction,
        which gives the direction of E-field polarisation"""
        def __get__(self):
            return (self.ray.E_vector.x,self.ray.E_vector.y,self.ray.E_vector.z)
        
        def __set__(self, v):
            self.ray.E_vector.x = v[0]
            self.ray.E_vector.y = v[1]
            self.ray.E_vector.z = v[2]
            
    property length:
        """The length of the ray. This is infinite in 
        unterminated rays"""
        def __get__(self):
            return self.ray.length
        
        def __set__(self, double v):
            self.ray.length = v
            
    property phase:
        """An additional phase-factor for the ray. At present, this handles the 'grating phase' factor
        generated by diffraction gratings. All other material surfaces leave this unchanged"""
        def __get__(self):
            return self.ray.phase
        
        def __set__(self, double v):
            self.ray.phase = v
            
    property accumulated_path:
        """The total *optical* path up to the start-point of this ray."""
        def __get__(self):
            return self.ray.accumulated_path
        
        def __set__(self, double v):
            self.ray.accumulated_path = v
            
    property wavelength_idx:
        """The wavelength of the ray in vacuum, in microns"""
        def __get__(self):
            return self.ray.wavelength_idx
        
        def __set__(self, unsigned int v):
            self.ray.wavelength_idx = v
            
    property termination:
        """the end-point of the ray (read only)
        """
        def __get__(self):
            cdef vector_t end
            cdef float length
            if self.ray.length > self.max_length:
                length = self.max_length
            else:
                length = self.ray.length
            end = addvv_(self.ray.origin, multvs_(self.ray.direction, 
                                    length))
            return (end.x, end.y, end.z)
        
    property refractive_index:
        """complex refractive index through which
        this ray is propagating"""
        def __get__(self):
            return complex(self.ray.refractive_index.real,
                            self.ray.refractive_index.imag)
        def __set__(self, v):
            self.ray.refractive_index.real = v.real
            self.ray.refractive_index.imag = v.imag
            
    property E1_amp:
        """Complex amplitude of the electric field polarised
        parallel to the E_vection."""
        def __get__(self):
            return complex(self.ray.E1_amp.real,
                            self.ray.E1_amp.imag)
        def __set__(self, v):
            self.ray.E1_amp.real = v.real
            self.ray.E1_amp.imag = v.imag
            
    property E2_amp:
        """Complex amplitude of the electric field polarised
        perpendicular to the E_vection"""
        def __get__(self):
            return complex(self.ray.E2_amp.real,
                            self.ray.E2_amp.imag)
        def __set__(self, v):
            self.ray.E2_amp.real = v.real
            self.ray.E2_amp.imag = v.imag
            
    property parent_idx:
        """Index of the parent ray in parent RayCollection
        """
        def __get__(self):
            return self.ray.parent_idx
        
        def __set__(self, unsigned int v):
            self.ray.parent_idx = v
            
    property end_face_idx:
        """Index of the terminating face, in the global
        face list (created for each tracing operation)
        """
        def __get__(self):
            return self.ray.end_face_idx
        
        def __set__(self, unsigned int v):
            self.ray.end_face_idx = v
            
    property ray_type_id:
        """Used to distinguish rays created by reflection vs transmission or some other mechanism.
        Transmission->0, Reflection->1"""
        def __get__(self):
            return self.ray.ray_type_id
        
        def __set__(self, unsigned int v):
            self.ray.ray_type_id = v
            
    property power:
        """Optical power for the ray"""
        def __get__(self):
            return ray_power_(self.ray)
        
    property amplitude:
        """E field amplitude"""
        def __get__(self):
            cdef ray_t ray=self.ray
            return sqrt(ray.E1_amp.real**2 + ray.E1_amp.imag**2 +\
                        ray.E2_amp.real**2 + ray.E2_amp.imag**2)
            
    property jones_vector:
        """Jones polarisation vector expressed as a tuple (alpha, beta)
        where alpha and beta are complex"""
        def __get__(self):
            cdef double amp = self.amplitude
            return (self.E1_amp/amp, self.E2_amp/amp)
        
    property E_left:
        def __get__(self):
            E1_amp = self.E1_amp
            E2_amp = self.E2_amp
            return (E1_amp + 1.0j*E2_amp)/sqrt(2)
        
    property E_right:
        def __get__(self):
            E1_amp = self.E1_amp
            E2_amp = self.E2_amp
            return (E1_amp - 1.0j*E2_amp)/sqrt(2)
        
    property ellipticity:
        """Provide the ratio of power in the RH circular polarisation
        to the LH circular polarisation. A value of zero indicates 
        linear polarisation. +1 indicate RH polarisation, -1 is
        LH polarisation. Or maybe the other way round."""
        def __get__(self):
            alpha, beta = self.jones_vector
            R = (alpha - 1j*beta)/root2
            L = (alpha + 1j*beta)/root2
            PR = R.real**2 + R.imag**2
            PL = L.real**2 + L.imag**2
            return (PR - PL)/(PR + PL)
        
    property major_minor_axes:
        """Find the vector giving the major and minor axes of the polarisation ellipse.
        For fully circularly polarised light, the current E_vector will be
        returned"""
        def __get__(self):
            cdef:
                vector_t E1_vector, E2_vector
                
            E2_vector = norm_(cross_(self.ray.direction, self.ray.E_vector))
            E1_vector = norm_(cross_(E2_vector, self.ray.direction))
            
            E1 = sqrt(self.ray.E1_amp.real**2 + self.ray.E1_amp.imag**2)
            E2 = sqrt(self.ray.E2_amp.real**2 + self.ray.E2_amp.imag**2)
            phi1 = atan2(self.ray.E1_amp.imag, self.ray.E1_amp.real)
            phi2 = atan2(self.ray.E2_amp.imag, self.ray.E2_amp.real)
            
            A = E1*E1*sin(2*phi1) + E2*E2*sin(2*phi2)
            B = E1*E1*cos(2*phi1) + E2*E2*cos(2*phi2)
            
            mag = sqrt(A*A + B*B)
            t1 = atan( (B - mag)/A )
            t2 = atan( (B + mag)/A )
            
            axis1 = addvv_(multvs_(E1_vector, E1*cos(phi1+t1)),
                           multvs_(E2_vector, E2*cos(phi2+t1)))
            
            axis2 = addvv_(multvs_(E1_vector, E1*cos(phi1+t2)),
                           multvs_(E2_vector, E2*cos(phi2+t2)))
            
            return ( (axis1.x, axis1.y, axis1.z), 
                     (axis2.x, axis2.y, axis2.z) )
            
        
    def project_E(self, *axis):
        """Rotate the E_vector onto the given axis, projecting
        E1_amp and E2_amp as necessary."""
        cdef:
            complex_t E1, E2
            vector_t v=set_v(axis)
            vector_t E_vector, E_vector2bar, E_vector2, direction
            double A, B
            
        E1 = self.ray.E1_amp
        E2 = self.ray.E2_amp
        direction = self.ray.direction
        
        #initial E_vector
        E_vector = norm_(cross_(cross_(direction, self.ray.E_vector), direction))
        #new E_vector
        E_vector2bar = norm_(cross_(direction, v))
        E_vector2 = norm_(cross_(E_vector2bar, direction))
        
        A = dotprod_(E_vector, E_vector2)
        B = dotprod_(E_vector, E_vector2bar)
        
        #print "A,B,AB^2:", A, B, A**2 + B**2
        
        self.ray.E1_amp.real = -(E2.real*B - E1.real*A)
        self.ray.E2_amp.real = (E2.real*A + E1.real*B)
        
        self.ray.E1_amp.imag = -(E2.imag*B - E1.imag*A)
        self.ray.E2_amp.imag = E2.imag*A + E1.imag*B
        
        self.ray.E_vector = E_vector2
        
        
cdef class Gausslet:
    def __cinit__(self, **kwds):
        for k in kwds:
            setattr(self, k, kwds[k])
            
    def __repr__(self):
        return "Gausslet(o=%s, d=%s)"%(str(self.base_ray.origin),
                                            str(self.base_ray.direction))
        
    property base_ray:
        def __get__(self):
            cdef:
                Ray out = Ray()
                
            out.ray = self.gausslet.base_ray
            return out
        
        def __set__(self, Ray r):
            self.gausslet.base_ray = r.ray
            
    property parabasal_rays:
        def __get__(self):
            cdef:
                ParabasalRay p
                int i
                tuple out
                
            out = tuple(ParabasalRay() for i in range(6))
            for i in range(6):
                p = out[i]
                p.ray = self.gausslet.para[i]
            return out
        
        def __set__(self, tuple paras):
            for i in range(6):
                self.gausslet.para[i] = paras[i].ray
                
                
cdef class RayArrayView:
    """An abstract class to provide the API for ray_t member access from python / numpy"""
    cdef void set_ray_c(self, unsigned long i, ray_t ray):
        return
    
    cdef ray_t get_ray_c(self, unsigned long i):
        cdef:
            ray_t r
        return r
    
    cdef unsigned long get_n_rays(self):
        return 0
    
    def __getitem__(self, size_t idx):
        cdef Ray r
        if idx >= self.get_n_rays():
            raise IndexError("Requested index %d from a size %d array"%(idx, self.n_rays))
        r = Ray()
        r.ray = self.get_ray_c(idx)
        return r
    
    def __setitem__(self, size_t idx, Ray r):
        if idx >= self.get_n_rays():
            raise IndexError("Attempting to set index %d from a size %d array"%(idx, self.n_rays))
        self.set_ray_c(idx, r.ray)
        
    def get_ray_list(self):
        """Returns the contents of this RayCollection as a list of Rays
        """
        cdef size_t i
        cdef list ray_list = []
        cdef Ray r
        for i in range(self.get_n_rays()):
            r = Ray()
            r.ray = self.get_ray_c(i)
            ray_list.append(r)
        return ray_list
    
    property origin:
        def __get__(self):
            cdef:
                unsigned long n=self.get_n_rays()
                np_.ndarray out = np.empty((n,3), dtype='d')
                size_t i
                vector_t v
            for i in xrange(n):
                v = self.get_ray_c(i).origin
                out[i,0] = v.x
                out[i,1] = v.y
                out[i,2] = v.z
            return out
        
    property direction:
        def __get__(self):
            cdef:
                unsigned long n=self.get_n_rays()
                np_.ndarray out = np.empty((n,3), dtype='d')
                size_t i
                vector_t v
            for i in xrange(n):
                v = self.get_ray_c(i).direction
                out[i,0] = v.x
                out[i,1] = v.y
                out[i,2] = v.z
            return out
        
    property normal:
        def __get__(self):
            cdef:
                unsigned long n=self.get_n_rays()
                np_.ndarray out = np.empty((n,3), dtype='d')
                size_t i
                vector_t v
            for i in xrange(n):
                v = self.get_ray_c(i).normal
                out[i,0] = v.x
                out[i,1] = v.y
                out[i,2] = v.z
            return out
        
    property E_vector:
        def __get__(self):
            cdef:
                unsigned long n=self.get_n_rays()
                np_.ndarray out = np.empty((n,3), dtype='d')
                size_t i
                vector_t v
            for i in xrange(n):
                v = self.get_ray_c(i).E_vector
                out[i,0] = v.x
                out[i,1] = v.y
                out[i,2] = v.z
            return out
        
    property refractive_index:
        def __get__(self):
            cdef:
                unsigned long N=self.get_n_rays()
                np_.ndarray out = np.empty(N, dtype=np.complex128)
                np_.ndarray rr,ii
                size_t i
                complex_t n
            rr = out.real
            ii = out.imag
            for i in xrange(N):
                n = self.get_ray_c(i).refractive_index
                rr[i] = n.real
                ii[i] = n.imag
            return out
        
    property E1_amp:
        def __get__(self):
            cdef:
                unsigned long N=self.get_n_rays()
                np_.ndarray out = np.empty(N, dtype=np.complex128)
                np_.ndarray rr,ii
                size_t i
                complex_t n
            rr = out.real
            ii = out.imag
            for i in xrange(N):
                n = self.get_ray_c(i).E1_amp
                rr[i] = n.real
                ii[i] = n.imag
            return out
        
    property E2_amp:
        def __get__(self):
            cdef:
                unsigned long N=self.get_n_rays()
                np_.ndarray out = np.empty(N, dtype=np.complex128)
                np_.ndarray rr,ii
                size_t i
                complex_t n
            rr = out.real
            ii = out.imag
            for i in xrange(N):
                n = self.get_ray_c(i).E2_amp
                rr[i] = n.real
                ii[i] = n.imag
            return out
        
    property length:
        def __get__(self):
            cdef:
                unsigned long N=self.get_n_rays()
                np_.ndarray out = np.empty(N, dtype='d')
                size_t i
                double v
            for i in xrange(N):
                v = self.get_ray_c(i).length
                out[i] = v
            return out
        
    property phase:
        def __get__(self):
            cdef:
                unsigned long N=self.get_n_rays()
                np_.ndarray out = np.empty(N, dtype='d')
                size_t i
                double v
            for i in xrange(N):
                v = self.get_ray_c(i).phase
                out[i] = v
            return out
        
    property accumulated_path:
        def __get__(self):
            cdef:
                unsigned long N=self.get_n_rays()
                np_.ndarray out = np.empty((N,), dtype='d' )
                size_t i
            for i in xrange(N):
                out[i] = self.get_ray_c(i).accumulated_path
            return out
        
    property wavelength_idx:
        def __get__(self):
            cdef:
                unsigned long N=self.get_n_rays()
                np_.ndarray out = np.empty(N, dtype=np.uint32)
                size_t i
                double v
            for i in xrange(N):
                v = self.get_ray_c(i).wavelength_idx
                out[i] = v
            return out
        
    property parent_idx:
        def __get__(self):
            cdef:
                unsigned long N=self.get_n_rays()
                np_.ndarray out = np.empty(N, dtype=np.uint32)
                size_t i
                unsigned int v
            for i in xrange(N):
                v = self.get_ray_c(i).parent_idx
                out[i] = v
            return out
        
    property end_face_idx:
        def __get__(self):
            cdef:
                unsigned long N=self.get_n_rays()
                np_.ndarray out = np.empty(N, dtype=np.uint32)
                size_t i
                unsigned int v
            for i in xrange(N):
                v = self.get_ray_c(i).end_face_idx
                out[i] = v
            return out
        
    property ray_type_id:
        def __get__(self):
            cdef:
                unsigned long N=self.get_n_rays()
                np_.ndarray out = np.empty(N, dtype=np.uint32)
                size_t i
                unsigned int v
            for i in xrange(N):
                v = self.get_ray_c(i).ray_type_id
                out[i] = v
            return out
        
    property termination:
        def __get__(self):
            cdef:
                unsigned long N=self.get_n_rays()
                np_.ndarray out = np.empty((N,3), dtype='d')
                size_t i
                vector_t v
                ray_t r
            for i in xrange(N):
                r = self.get_ray_c(i)
                v = addvv_(r.origin, 
                           multvs_(r.direction, 
                                    r.length))
                out[i,0] = v.x
                out[i,1] = v.y
                out[i,2] = v.z
            return out
    
    

cdef class RayCollection(RayArrayView):
    """A list-like collection of ray_t objects.
    
    The RayCollection is the primary data-structure used in the ray-tracing operation. 
    
    The RayCollection is of variable length, in that it can grow as individual rays are added to it.
    Internally, the memory allocated to the array of ray_t structures is re-allocated to increase
    its capacity.
    """
    
    def __cinit__(self, size_t max_size):
        self.rays = <ray_t*>malloc(max_size*sizeof(ray_t))
        self.n_rays = 0
        self.max_size = max_size
        self._mtime = 0.0
        
    def __dealloc__(self):
        free(self.rays)
        
    def __len__(self):
        return self.n_rays
    
    cdef ray_t get_ray_c(self, unsigned long i):
        return self.rays[i]
    
    cdef void set_ray_c(self, unsigned long i, ray_t ray):
        self.rays[i] = ray
        
    cdef unsigned long get_n_rays(self):
        return self.n_rays
        
    cdef add_ray_c(self, ray_t r):
        if self.n_rays == self.max_size:
            if self.max_size == 0:
                self.max_size = 1
            else:
                self.max_size *= 2
            self.rays = <ray_t*>realloc(self.rays, self.max_size*sizeof(ray_t))
        self.rays[self.n_rays] = r
        self.n_rays += 1
        
    cdef void reset_length_c(self, double max_length):
        cdef:
            size_t i
        for i in range(self.n_rays):
            self.rays[i].length = max_length
            
    def reset_length(self, double max_length=INF):
        """Sets the length of all rays in this RayCollection to Infinity
        """
        self.reset_length_c(max_length)
        
    def add_ray(self, Ray r):
        """Adds the given Ray instance to this collection
        """
        self.add_ray_c(r.ray)
        
    def add_ray_list(self, list rays):
        """Adds the given list of Rays to this collection
        """
        cdef long int i
        for i in range(len(rays)):
            if not isinstance(rays[i], Ray):
                raise TypeError("ray list contains non-Ray instance at index %d"%i)
        for i in range(len(rays)):
            self.add_ray_c((<Ray>rays[i]).ray)
        
    def clear_ray_list(self):
        """Empties this RayCollection (by setting the count to zero)
        """
        self.n_rays = 0
        
    def __iter__(self):
        return RayCollectionIterator(self)
    
    def copy_as_array(self):
        """Returns the contents of this RayCollection as a numpy array
        (the data is always copied).
        """
        cdef np_.ndarray out = np.empty(self.n_rays, dtype=ray_dtype)
        memcpy(<np_.float64_t *>out.data, self.rays, self.n_rays*sizeof(ray_t))
        return out
    
    property wavelengths:
        def __get__(self):
            return np.asarray(self._wavelengths)
        
        def __set__(self, wl_list):
            self._wavelengths = np.ascontiguousarray(wl_list, dtype=np.double)
    
    cdef double get_mtime(self, unsigned long guard):
        cdef:
            double pmtime
            
        if self._parent is None:
            return self._mtime
        if guard == id(self):
            return self._mtime
        
        pmtime = self._parent.get_mtime(guard)
        if pmtime > self._mtime:
            self._neighbours = None
            return pmtime
        else:
            return self._mtime
        
    cdef void _eval_neighbours(self, int[:,:] pnb):
        cdef:
            unsigned long i
            int j, pidx, child_nb, rtype
            int[:,:] rmap
            unsigned int nparent=self._parent.n_rays
            
        if pnb is None:
            return
        
        rmap = np.full( (nparent, 2), -1 ,dtype=np.int32)
        nb = np.full( (self.n_rays, pnb.shape[1]), -1, dtype=np.int32)
        
        for i in range(self.n_rays):
            rtype = self.rays[i].ray_type_id & REFL_RAY
            pidx = self.rays[i].parent_idx
            rmap[pidx, rtype] = i
            
        for i in range(self.n_rays):
            rtype = self.rays[i].ray_type_id & REFL_RAY
            pidx = self.rays[i].parent_idx
            for j in range(pnb.shape[1]):
                child_nb = pnb[pidx,j]
                if child_nb >= 0:
                    nb[i,j] = rmap[child_nb,rtype]
                    
        self._neighbours = nb
        
        
    property neighbours:
        def __get__(self):
            if self._parent is None:
                if self._neighbours is None:
                    return None
                else:
                    return np.asarray(self._neighbours)
            else:
                pmtime = self._parent.get_mtime(id(self))
                if self._mtime >= pmtime:
                    if self._neighbours is not None:
                        return np.asarray(self._neighbours)
                self._eval_neighbours(self._parent.neighbours)
                self._mtime = time.monotonic()
                return np.asarray(self._neighbours)
                
        def __set__(self, int[:,:] nb):
            self._neighbours = nb
            self._mtime = time.monotonic()
            
    property parent:
        def __get__(self):
            return self._parent
        
        def __set__(self, RayCollection rc):
            self._parent = rc
            self._neighbours = None
            self._wavelengths = rc._wavelengths
    
    @classmethod
    def from_array(cls, np_.ndarray data):
        """Creates a new RayCollection from the given numpy array. The array
        dtype should be a ctracer.ray_dtype. The data is copied into the 
        RayCollection
        """
        cdef int size=data.shape[0]
        cdef RayCollection rc = RayCollection(size)
        assert data.dtype is ray_dtype
        data = np.ascontiguousarray(data)
        memcpy(rc.rays, <np_.float64_t *>data.data, size*sizeof(ray_t))
        rc.n_rays = size
        return rc
    
    
cdef class GaussletBaseRayView(RayArrayView):
    def __cinit__(self, GaussletCollection owner):
        self.owner = owner

    cdef void set_ray_c(self, unsigned long i, ray_t ray):
        self.owner.rays[i].base_ray = ray
    
    cdef ray_t get_ray_c(self, unsigned long i):
        return self.owner.rays[i].base_ray
    
    cdef unsigned long get_n_rays(self):
        return self.owner.n_rays
    
    def __len__(self):
        return self.owner.n_rays
    
    def copy_as_array(self):
        cdef:
            unsigned int i, N = self.get_n_rays()
            np_.ndarray out = np.empty((N,), dtype=ray_dtype)
            ray_t * rays ###A bug in Cython memory views means we cannot use ray_t[:]
            
        rays = <ray_t *>(out.data)
        for i in range(N):
            rays[i] = self.owner.rays[i].base_ray
        return out
    
    
cdef class GaussletCollection:
    """A list-like collection of ray_t objects.
    
    The RayCollection is the primary data-structure used in the ray-tracing operation. 
    
    The RayCollection is of variable length, in that it can grow as individual rays are added to it.
    Internally, the memory allocated to the array of ray_t structures is re-allocated to increase
    its capacity.
    """
    
    def __cinit__(self, size_t max_size):
        self.rays = <gausslet_t*>malloc(max_size*sizeof(gausslet_t))
        self.n_rays = 0
        self.max_size = max_size
        
    def __dealloc__(self):
        free(self.rays)
        
    def __len__(self):
        return self.n_rays
    
    property parent:
        def __get__(self):
            return self._parent
        
        def __set__(self, GaussletCollection gc):
            self._parent = gc
            self._wavelengths = gc._wavelengths
        
    cdef add_gausslet_c(self, gausslet_t r):
        if self.n_rays == self.max_size:
            if self.max_size == 0:
                self.max_size = 1
            else:
                self.max_size *= 2
            self.rays = <gausslet_t*>realloc(self.rays, self.max_size*sizeof(gausslet_t))
        self.rays[self.n_rays] = r
        self.n_rays += 1
        
    def add_gausslet(self, Gausslet r):
        """Adds the given Ray instance to this collection
        """
        self.add_gausslet_c(r.gausslet)
        
    def add_gausslet_list(self, list rays):
        """Adds the given list of Rays to this collection
        """
        cdef long int i
        for i in range(len(rays)):
            if not isinstance(rays[i], Gausslet):
                raise TypeError("ray list contains non-Gausslet instance at index %d"%i)
        for i in range(len(rays)):
            self.add_ray_c((<Gausslet>rays[i]).gausslet)
            
    cdef void reset_length_c(self, double max_length):
        cdef:
            size_t i, j
        for i in range(self.n_rays):
            self.rays[i].base_ray.length = max_length
            for j in range(6):
                self.rays[i].para[j].length = max_length
        
            
    def reset_length(self, double max_length=INF):
        """Sets the length of all rays in this RayCollection to Infinity
        """
        self.reset_length_c(max_length)
        
    def clear_ray_list(self):
        """Empties this RayCollection (by setting the count to zero)
        """
        self.n_rays = 0
        
    def get_gausslet_list(self):
        """Returns the contents of this RayCollection as a list of Rays
        """
        cdef size_t i
        cdef list ray_list = []
        cdef Gausslet r
        for i in range(self.n_rays):
            r = Gausslet()
            r.gausslet = self.rays[i]
            ray_list.append(r)
        return ray_list
    
    def __getitem__(self, size_t idx):
        cdef Gausslet r
        if idx >= self.n_rays:
            raise IndexError("Requested index %d from a size %d array"%(idx, self.n_rays))
        r = Gausslet()
        r.gausslet = self.rays[idx]
        return r
    
    def __setitem__(self, size_t idx, Gausslet r):
        if idx >= self.n_rays:
            raise IndexError("Attempting to set index %d from a size %d array"%(idx, self.n_rays))
        self.rays[idx] = r.gausslet
        
    def __iter__(self):
        return GaussletCollectionIterator(self)
    
    def copy_as_array(self):
        """Returns the contents of this RayCollection as a numpy array
        (the data is always copied).
        """
        cdef np_.ndarray out = np.empty(self.n_rays, dtype=gausslet_dtype)
        memcpy(<np_.float64_t *>out.data, self.rays, self.n_rays*sizeof(gausslet_t))
        return out
    
    def extend(self, GaussletCollection gc):
        self.extend_c(gc)
        
    cdef void extend_c(self, GaussletCollection gc):
        if (self.n_rays + gc.n_rays) > self.max_size:
            self.max_size = (self.n_rays*2 + gc.n_rays)
            self.rays = <gausslet_t*>realloc(self.rays, self.max_size*sizeof(gausslet_t))
        memcpy(self.rays + self.n_rays, gc.rays, gc.n_rays*sizeof(gausslet_t))
        self.n_rays += gc.n_rays
    
    @classmethod
    def from_array(cls, np_.ndarray data):
        """Creates a new RayCollection from the given numpy array. The array
        dtype should be a ctracer.ray_dtype. The data is copied into the 
        RayCollection
        """
        cdef: 
            int size=data.shape[0]
            GaussletCollection rc = GaussletCollection(size)
            
        if data.dtype != gausslet_dtype:
            raise ValueError("Array must have gausslet_dtype dtype")
        
        memcpy(rc.rays, <np_.float64_t *>data.data, size*sizeof(gausslet_t))
        rc.n_rays = size
        return rc
    
    @classmethod
    def from_rays(cls, np_.ndarray data):
        cdef: 
            int i, j, size=data.shape[0]
            GaussletCollection rc = GaussletCollection(size)
            gausslet_t *gc
            ray_t *ray
            para_t *p
            ray_t *_data = <ray_t *>data.data
            
        if data.dtype != ray_dtype:
            raise ValueError("Array must have gausslet_dtype dtype")
            
        for i in range(size):
            gc = rc.rays+i
            ray = &(_data[i])
            gc.base_ray = ray[0]
            for j in range(6):
                p = gc.para + j
                p.origin = ray.origin
                p.direction = ray.direction
                p.normal = ray.normal
                p.length = ray.length
        rc.n_rays = size
        return rc
    
    property lagrange_invariant:
        def __get__(self):
            cdef:
                int i,j
                unsigned long k, N = self.n_rays
                np_.ndarray out = np.empty((N,), dtype='d' )
                double[:] _out = out
                double v=0
                gausslet_t g
                ray_t r
                para_t p
                vector_t[6] h
                vector_t[6] u
                vector_t axis1, axis2, o
                double[:] wavelen = self.wavelengths
                
                
            for k in range(N):
                g = self.rays[k]
                r = g.base_ray
                axis1 = norm_(r.E_vector)
                axis2 = cross_(axis1, r.direction)
                
                for i in range(6):
                    p = g.para[i]
                    o = subvv_(p.origin, r.origin)
                    h[i].x = dotprod_(o, axis1)
                    h[i].y = dotprod_(o, axis2)
                    h[i].z = 0.0
                    u[i].x = dotprod_(p.direction, axis1)
                    u[i].y = dotprod_(p.direction, axis2)
                    u[i].z = 0.0
                    
                v = 0
                v += (dotprod_(h[0],u[5]) - dotprod_(h[5],u[0]))**2
                v += (dotprod_(h[1],u[2]) - dotprod_(h[2],u[1]))**2
                v += (dotprod_(h[3],u[4]) - dotprod_(h[4],u[3]))**2
                v += (dotprod_(h[1],u[4]) - dotprod_(h[4],u[1]))**2
                v += (dotprod_(h[0],u[3]) - dotprod_(h[3],u[0]))**2
                v += (dotprod_(h[2],u[5]) - dotprod_(h[5],u[2]))**2
                
                _out[k] = 1000*sqrt(v/6)/ wavelen[r.wavelength_idx]
            return out
        
    def project_to_plane(self, origin, direction):
        """
        Project the rays in the collection onto the intersection with the given plane,
        defined by an origin point on the plane and the plane normal vector.
        """
        cdef:
            vector_t o,d
            unsigned int i, j
            gausslet_t *gc
            ray_t *ray
            para_t *para
            double a
            #double complex[:] wl = self.wavelengths
            
        o.x, o.y, o.z = origin
        d.x, d.y, d.z = direction
        d = norm_(d)
        
        for i in range(self.n_rays):
            gc = &(self.rays[i])
            ray = &(gc.base_ray)
            a = dotprod_(subvv_(o, ray.origin),d) / dotprod_(ray.direction, d)
            ray.origin = addvv_(ray.origin, multvs_(ray.direction,a))
            ray.accumulated_path += ray.refractive_index.real * a
            ### Handle absorption along ray
            ### Am ignoring this for now. FIXME
            for j in range(6):
                para = &(gc.para[j])
                a = dotprod_(subvv_(o, para.origin),d) / dotprod_(para.direction, d)
                para.origin = addvv_(para.origin, multvs_(para.direction,a))
                
    def scale_amplitude(self, double complex scale):
        cdef:
            unsigned long i=self.n_rays
            
        for i in range(self.n_rays):
            self.rays[i].base_ray.E1_amp *= scale
            self.rays[i].base_ray.E2_amp *= scale
    
    def config_parabasal_rays(self, double[:] wavelength_list, double radius, double working_dist):
        """
        Initialise the parabasal rays for a symmetric (i.e. circular) modes, 
        using the base_ray data for wavelength, and the given beam waist 1/e^2 radius.
        'working_dist' indicates the distance from the base_ray origin to the centre 
        of the gaussian beam waist. Negative values imply a beam waist before the origin. 
        'radius' is given in mm.
        """
        cdef:
            int i,j
            gausslet_t *gc
            double theta0, denom, angle
            vector_t o, d, d1, d2, base_d, da, db
            
        for i in range(self.n_rays):
            gc = self.rays+i
            base_d = norm_(gc.base_ray.direction)
            if base_d.x > base_d.y:
                o.x=0.0
                o.y=1.0
                o.z=0.0
            else:
                o.x=1.0
                o.y=0.0
                o.z=0.0
            d1 = norm_(cross_(base_d, o))
            d2 = norm_(cross_(base_d, d1))
            ### Divergence of the gausslet
            theta0 = wavelength_list[gc.base_ray.wavelength_idx]/(M_PI*radius*1000.0)
            
            for j in range(0,6,2):
                angle = (j*2*M_PI/6)# + i*(2*M_PI)/self.n_rays
                o = addvv_(multvs_(d1, radius*cos(angle)),
                                multvs_(d2, radius*sin(angle)))
                o = addvv_(o, gc.base_ray.origin)
                o = addvv_(o, multvs_(base_d, working_dist))
                ###direction of ray
                d = addvv_(multvs_(d1, -theta0*sin(angle)),
                            multvs_(d2, theta0*cos(angle)))
                da = addvv_(base_d, d)
                db = subvv_(base_d, d)
                
                gc.para[j].direction = norm_(da)
                gc.para[j].origin = subvv_(o, multvs_(da,working_dist))
                
                gc.para[j+1].direction = norm_(db)
                gc.para[j+1].origin = subvv_(o, multvs_(db,working_dist))
                
                gc.para[j].normal = gc.base_ray.normal
                gc.para[j+1].normal = gc.base_ray.normal 
                gc.para[j].length = gc.base_ray.length
                gc.para[j+1].length = gc.base_ray.length
                
    property base_rays:
        def __get__(self):
            return GaussletBaseRayView(self)
        
    property total_power:
        def __get__(self):
            cdef:
                unsigned long i
                double pwr=0.0
                ray_t *ray
                double n
                
            for i in range(self.n_rays):
                ray = &(self.rays[i].base_ray)
                n = ray.refractive_index.real
                pwr += (ray.E1_amp.real * ray.E1_amp.real)*n
                pwr += (ray.E1_amp.imag * ray.E1_amp.imag)*n
                pwr += (ray.E2_amp.real * ray.E2_amp.real)*n
                pwr += (ray.E2_amp.imag * ray.E2_amp.imag)*n
            return pwr
                
    property wavelengths:
        def __get__(self):
            return np.asarray(self._wavelengths)
        
        def __set__(self, wl_list):
            self._wavelengths = np.asarray(wl_list, dtype=np.double)
        
    property para_origin:
        def __get__(self):
            cdef:
                np_.ndarray out = np.empty((self.n_rays,6,3), dtype='d')
                size_t i, j
                vector_t v
            for i in xrange(self.n_rays):
                for j in xrange(6):
                    v = self.rays[i].para[j].origin
                    out[i,j,0] = v.x
                    out[i,j,1] = v.y
                    out[i,j,2] = v.z
            return out
        
    property para_direction:
        def __get__(self):
            cdef:
                np_.ndarray out = np.empty((self.n_rays,6,3), dtype='d')
                size_t i, j
                vector_t v
            for i in xrange(self.n_rays):
                for j in range(6):
                    v = self.rays[i].para[j].direction
                    out[i,j,0] = v.x
                    out[i,j,1] = v.y
                    out[i,j,2] = v.z
            return out
        
    property para_normal:
        def __get__(self):
            cdef:
                np_.ndarray out = np.empty((self.n_rays,6,3), dtype='d')
                size_t i, j
                vector_t v
            for i in xrange(self.n_rays):
                for j in range(6):
                    v = self.rays[i].para[j].normal
                    out[i,j,0] = v.x
                    out[i,j,1] = v.y
                    out[i,j,2] = v.z
            return out
        
    property para_termination:
        def __get__(self):
            cdef:
                np_.ndarray aout = np.empty((self.n_rays,6,3), dtype='d')
                double [:,:,:] out = aout
                size_t i,j
                vector_t v
                para_t *p
            for i in xrange(self.n_rays):
                for j in range(1,7):
                    p = self.rays[i].para + j
                    v = addvv_(p.origin, 
                           multvs_(p.direction, 
                                    p.length))
                    out[i,j,0] = v.x
                    out[i,j,1] = v.y
                    out[i,j,2] = v.z
            return aout
        
    
cdef class InterfaceMaterial(object):
    """Abstract base class for objects describing
    the materials characterics of a Face
    """
    
    def __cinit__(self):
        self.wavelengths = np.array([], dtype=np.double)
    
    cdef void eval_child_ray_c(self, ray_t *old_ray, 
                                unsigned int ray_idx, 
                                vector_t p, 
                                orientation_t orient,
                                RayCollection new_rays):
        pass
    
    cdef para_t eval_parabasal_ray_c(self, ray_t *base_ray, 
                                     vector_t direction, #incoming ray direction
                                   vector_t point, #position of intercept
                                   orientation_t orient,
                                   unsigned int ray_type_id, #bool, if True, it's a reflected ray
                                   ):
        cdef:
            vector_t cosThetaNormal, reflected, normal
            para_t para_out
            double cosTheta
        
        normal = norm_(orient.normal)
        if ray_type_id & REFL_RAY:
            cosTheta = dotprod_(normal, direction)
            cosThetaNormal = multvs_(normal, cosTheta)
            reflected = subvv_(direction, multvs_(cosThetaNormal, 2))
            para_out.direction = reflected
        else:
            para_out.direction = direction
        para_out.origin = point
        para_out.normal = normal
        para_out.length = INF
        return para_out
    
    def is_decomp_material(self):
        return False
    
    cdef void eval_decomposed_rays_c(self, GaussletCollection child_rays):
        ### This function is called at the end of tracing to comput additional rays
        ### due to mode decomposition.
        return
    
    def eval_child_ray(self, Ray old_ray, ray_idx, point, 
                        normal, tangent, RayCollection new_rays):
        cdef:
            vector_t p
            orientation_t n
            Ray out=Ray()
            unsigned int idx
        
        p = set_v(point)
        n.normal = set_v(normal)
        n.tangent = set_v(tangent)
        self.eval_child_ray_c(&old_ray.ray, ray_idx, 
                                        p, n, new_rays)
        
    def eval_parabasal_ray(self, Ray base_ray, direction, point, normal, tangent, reflect=False):
        cdef:
            vector_t d=set_v(direction), p = set_v(point)
            orientation_t n
            ParabasalRay out = ParabasalRay()
            unsigned int ray_type=1 if reflect else 0
            
        n.normal = set_v(normal)
        n.tangent = set_v(tangent)
        out.ray = self.eval_parabasal_ray_c(&base_ray.ray, d, p, n, ray_type)
        return out
            
        
    property wavelengths:
        def __set__(self, double[:] wavelengths):
            self._wavelengths = wavelengths
            self.on_set_wavelengths()
            
        def __get__(self):
            return self._wavelengths
        
    cdef on_set_wavelengths(self):
        pass
    
    
cdef class Shape:        
    cdef bint point_inside_c(self, double x, double y):
        return 1
    
    def point_inside(self, double x, double y):
        return self.point_inside_c(x,y) 
    
    
cdef class Distortion:
    """A abstract base class to represents distortions on a face, a z-offset 
    as a function of (x,y).
    """
    cdef vector_t z_offset_and_gradient_c(self, double x, double y) nogil:
        """The z-axis surface sag is returned as the z-component 
        of the output vector. The x- and y-components of the surface
        gradient are placed in the x- and y- components of vector.
        I.e. the returned vector is
            (dz(x,y)/dx, dz(x,y)/dy, z(x,y))
        """
        cdef:
            vector_t p
        p.x=0.0
        p.y=0.0
        p.z=0.0
        return p
    
    cdef double z_offset_c(self, double x, double y) nogil:
        return 0.0
    
    def z_offset_and_gradient(self, double[:] x, double[:] y):
        cdef:
            vector_t v
            size_t i, n=len(x)
            double[:,:] out = np.empty((n, 3), 'd')
            
        if len(y) != n:
            raise ValueError("Both x and y must have the same length")
            
        for i in range(n):
            v = self.z_offset_and_gradient_c(x[i],y[i])
            out[i,0] = v.x
            out[i,1] = v.y
            out[i,2] = v.z
        return np.asarray(out)
    
    def z_offset(self, double[:] x, double[:] y):
        cdef:
            size_t i, n=len(x)
            double[:] out = np.empty((n,), 'd')
            
        if len(y) != n:
            raise ValueError("Both x and y must have the same length")
            
        for i in range(n):
            out[i] = self.z_offset_c(x[i],y[i])
        return np.asarray(out)
    
    
cdef class Face(object):
    
    params = []
    
    def __cinit__(self, owner=None, tolerance=0.0001, 
                        max_length=100, material=None, **kwds):
        self.name = "base Face class"
        self.tolerance = tolerance
        self.owner = owner
        self.max_length = max_length
        if isinstance(material, InterfaceMaterial):
            self.material = material
        else:
            from .cmaterials import PECMaterial
            self.material = PECMaterial()
        self.invert_normal = int(kwds.get('invert_normal', 0))
        
    
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        """returns the distance of the nearest valid intersection between 
        p1 and p2. p1 and p2 are in the local coordinate system
        """
        return 0
    
    def update(self):
        """Called to update the parameters from the owner
        to the Face
        """
        for name in self.params:
            v = getattr(self.owner, name)
            setattr(self, name, v)
    
    def intersect(self, p1, p2, int is_base_ray):
        cdef:
            vector_t p1_, p2_
            double dist
        
        p1_ = set_v(p1)
        p2_ = set_v(p2)
        dist = self.intersect_c(p1_, p2_, is_base_ray)
        return dist

    cdef vector_t compute_normal_c(self, vector_t p):
        return p
    
    cdef vector_t compute_tangent_c(self, vector_t p):
        cdef vector_t tangent
        tangent.x = 1.0
        tangent.y = 0.0
        tangent.z = 0.0
        return tangent
    
    def compute_normal(self, p):
        """Compute normal vector at a given point, in local
        face coordinates
        """
        cdef vector_t p_, n
        p_.x, p_.y, p_.z = p
        n = self.compute_normal_c(p_)
        return (n.x, n.y, n.z)
    
    def compute_tangent(self, p):
        """Compute the surface tangent at a given point,
        in face-local coordinates"""
        cdef vector_t tanget, p_
        
        p_.x, p_.y, p_.z = p
        tangent = self.compute_tangent_c(p_)
        return (tangent.x, tangent.y, tangent.z)
        


cdef class FaceList(object):
    """A group of faces which share a transform"""
    def __cinit__(self, owner=None):
        self.transform = Transform()
        self.inverse_transform = Transform()
        self.owner = owner
        
    cpdef void sync_transforms(self):
        """sets the transforms from the owner's VTKTransform
        """
        try:
            trans = self.owner.transform
        except AttributeError:
            print("NO OWNER", self.owner)
            return
        m = trans.matrix
        rot = [[m.get_element(i,j) for j in xrange(3)] for i in xrange(3)]
        dt = [m.get_element(i,3) for i in xrange(3)]
        #print "TRANS", rot, dt
        self.transform = Transform(rotation=rot, translation=dt)
        inv_trans = trans.linear_inverse
        m = inv_trans.matrix
        rot = [[m.get_element(i,j) for j in xrange(3)] for i in xrange(3)]
        dt = [m.get_element(i,3) for i in xrange(3)]
        self.inverse_transform = Transform(rotation=rot, translation=dt)
        
    property transform:
        def __set__(self, Transform t):
            self.trans = t.trans
           
        def __get__(self):
            cdef Transform t=Transform()
            t.trans = self.trans
            return t
        
    property inverse_transform:
        def __set__(self, Transform t):
            self.inv_trans = t.trans
           
        def __get__(self):
            cdef Transform t=Transform()
            t.trans = self.inv_trans
            return t
        
    def __getitem__(self, intidx):
        return self.faces[intidx]
        
     
    cdef int intersect_c(self, ray_t *ray, vector_t ray_end):
        """Finds the face with the nearest intersection
        point, for the ray defined by the two input points,
        P1 and P2 (in global coords).
        """
        cdef:
            vector_t p1 = transform_c(self.inv_trans, ray.origin)
            vector_t p2 = transform_c(self.inv_trans, ray_end)
            list faces=self.faces
            unsigned int i
            int all_idx=-1
            double dist
            Face face
        
        for i in xrange(len(faces)):
            face = faces[i]
            dist = face.intersect_c(p1, p2, 1)
            if face.tolerance < dist < ray.length:
                ray.length = dist
                all_idx = face.idx
                ray.end_face_idx = all_idx
        return all_idx
    
    def intersect(self, Ray r):
        cdef vector_t P1_
        cdef int idx
        
        P1_ = addvv_(r.ray.origin, multvs_(r.ray.direction, r.ray.length))
        idx = self.intersect_c(&r.ray, P1_)
        return idx
    
    cdef int intersect_para_c(self, para_t *ray, vector_t ray_end, Face face):
        cdef:
            vector_t p1 = transform_c(self.inv_trans, ray.origin)
            vector_t p2 = transform_c(self.inv_trans, ray_end)
            unsigned int i
            double dist
        
        dist = face.intersect_c(p1, p2, 0)
        #print("p1:", p1.x, p1.y, p1.z, "p2:", p2.x, p2.y, p2.z, "rlen:", ray.length, "dist:", dist)
        if face.tolerance < dist < ray.length:
            ray.length = dist
            return 0
        return -1
    
    def intersect_para(self, ParabasalRay r, Face face):
        cdef vector_t P1_
        cdef int idx
        
        P1_ = addvv_(r.ray.origin, multvs_(r.ray.direction, r.ray.length))
        idx = self.intersect_para_c(&r.ray, P1_, face)
        return idx
    
    cdef orientation_t compute_orientation_c(self, Face face, vector_t point):
        cdef orientation_t out
        
        point = transform_c(self.inv_trans, point)
        out.normal = face.compute_normal_c(point)
        out.tangent = face.compute_tangent_c(point)
        if face.invert_normal:
            out.normal = invert_(out.normal)
            out.tangent = invert_(out.tangent)
        out.normal = rotate_c(self.trans, out.normal)
        out.tangent = rotate_c(self.trans, out.tangent)
        return out
    
    def compute_orientation(self, Face face, point):
        cdef:
            vector_t p
            orientation_t o
        p = set_v(point)
        o = self.compute_orientation_c(face, p)
        return (o.normal.x, o.normal.y, o.normal.z), (o.tangent.x, o.tangent.y, o.tangent.z)
    

##################################
### Python module functions
##################################

cdef double ray_power_(ray_t ray):
    cdef double P1, P2
    
    P1 = (ray.E1_amp.real**2 + ray.E1_amp.imag**2)*ray.refractive_index.real
    P2 = (ray.E2_amp.real**2 + ray.E2_amp.imag**2)*ray.refractive_index.real
    ### We don't need to handle incident aspect area here as the ray already compensate for this
    #aspect = abs(dotprod(norm(ray.direction), normal))
    return (P1+P2) 
            
            
def select_ray_intersections(FaceList face_set, list ray_col_list):
    cdef:
        size_t i,j
        ray_t ray
        int idx, wl_offset=0
        vector_t point
        RayCollection rc, rc_out
        list wl_list=[]
        np_.int64_t[:] inverse
        
    rc_out = RayCollection(len(ray_col_list[0]))
        
    for j in range(len(ray_col_list)):
        rc = <RayCollection>ray_col_list[j]
        wl_list.append(rc.wavelengths)
        
        for i in range(rc.n_rays):
            ray = rc.rays[i] #need a copy of the ray to prevent mutating it in place
            point = addvv_(ray.origin, 
                                multvs_(ray.direction, 
                                        ray.length))
            
            idx = (<FaceList>face_set).intersect_c(&ray, point)
            if idx >= 0:
                ray.wavelength_idx += wl_offset
                rc_out.add_ray_c(ray)
        wl_offset += len(rc.wavelengths)
        
    reduced, inverse = np.unique(np.concatenate(wl_list), return_inverse=True)
    for i in range(len(rc_out)):
        idx = rc_out.rays[i].wavelength_idx
        rc_out.rays[i].wavelength_idx = inverse[idx]
        
    rc_out.wavelengths = reduced
    return rc_out
    
    
def select_gausslet_intersections(FaceList face_set, list ray_col_list):
    cdef:
        size_t i,j
        gausslet_t g
        ray_t *ray
        int idx, wl_offset=0
        vector_t point
        GaussletCollection rc, rc_out
        list wl_list=[]
        np_.int64_t[:] inverse
        
    rc_out = GaussletCollection(len(ray_col_list[0]))
        
    for j in range(len(ray_col_list)):
        rc = <GaussletCollection>ray_col_list[j]
        wl_list.append(rc.wavelengths)
        
        for i in range(rc.n_rays):
            g = rc.rays[i] #need a copy of the ray to prevent mutating it in place
            ray = &g.base_ray
            point = addvv_(ray.origin, 
                                multvs_(ray.direction, 
                                        ray.length))
            
            idx = (<FaceList>face_set).intersect_c(ray, point)
            if idx >= 0:
                ray.wavelength_idx += wl_offset
                rc_out.add_gausslet_c(g)
        wl_offset += len(rc.wavelengths)
        
    reduced, inverse = np.unique(np.concatenate(wl_list), return_inverse=True)
    for i in range(len(rc_out)):
        idx = rc_out.rays[i].base_ray.wavelength_idx
        rc_out.rays[i].base_ray.wavelength_idx = inverse[idx]
        
    rc_out.wavelengths = reduced
    return rc_out
    


cdef RayCollection trace_segment_c(RayCollection rays, 
                                    list face_sets, 
                                    list all_faces,
                                    list decomp_faces,
                                    float max_length):
    cdef:
        FaceList face_set #a FaceList
        Face face
        size_t i, j, n_sets=len(face_sets)
        vector_t point
        orientation_t orient
        int idx, nearest_set=-1, nearest_idx=-1
        ray_t *ray
        RayCollection new_rays
   
    #need to allocate the output rays here 
    new_rays = RayCollection(rays.n_rays)
    new_rays.parent = rays
    
    for i in range(rays.n_rays):
        ray = rays.rays + i
        ray.length = max_length
        ray.end_face_idx = -1
        nearest_idx=-1
        point = addvv_(ray.origin, 
                            multvs_(ray.direction, 
                                    max_length))
        #print "points", P1, P2
        for j in range(n_sets):
            face_set = face_sets[j]
            #intersect_c returns the face idx of the intersection, or -1 otherwise
            idx = (<FaceList>face_set).intersect_c(ray, point)
            if idx >= 0:
                nearest_set = j
                nearest_idx = idx
        if nearest_idx >= 0:
            #print "GET FACE", nearest.face_idx, len(all_faces)
            face = all_faces[nearest_idx]
            face.count += 1
            #print "ray length", ray.length
            point = addvv_(ray.origin, multvs_(ray.direction, ray.length))
            orient = (<FaceList>(face_sets[nearest_set])).compute_orientation_c(face, point)
            #print "s normal", normal
            (<InterfaceMaterial>(face.material)).eval_child_ray_c(ray, i, 
                                                    point,
                                                    orient,
                                                    new_rays
                                                    )
    return new_rays


def trace_segment(RayCollection rays, 
                    list face_sets, 
                    list all_faces,
                    max_length=100,
                    decomp_faces=[]):
    for fs in face_sets:
        fs.sync_transforms()
    return trace_segment_c(rays, face_sets, all_faces, decomp_faces, max_length)

def trace_gausslet(GaussletCollection rays, 
                    list face_sets, 
                    list all_faces,
                    max_length=100,
                    decomp_faces=[]):
    for fs in face_sets:
        (<FaceList>fs).sync_transforms()
    return trace_gausslet_c(rays, face_sets, all_faces, decomp_faces, max_length)


cdef GaussletCollection trace_gausslet_c(GaussletCollection gausslets, 
                                    list face_sets, 
                                    list all_faces,
                                    list decomp_faces,
                                    double max_length):
    cdef:
        FaceList face_set #a FaceList
        Face face
        size_t i, j, n_sets=len(face_sets), n_decomp = len(decomp_faces)
        vector_t point
        orientation_t orient
        int idx, nearest_set=-1, nearest_idx=-1
        gausslet_t *gausslet
        ray_t *ray
        GaussletCollection new_gausslets
        RayCollection child_rays
   
    #need to allocate the output rays here 
    new_gausslets = GaussletCollection(gausslets.n_rays)
    new_gausslets.parent = gausslets
    
    child_rays = RayCollection(2)
    
    for i in range(gausslets.n_rays):
        gausslet = gausslets.rays + i
        ray = &gausslet.base_ray
        ray.end_face_idx = -1
        nearest_idx=-1
        point = addvv_(ray.origin, 
                            multvs_(ray.direction, 
                                    max_length))
        #print "points", P1, P2
        for j in range(n_sets):
            face_set = face_sets[j]
            #intersect_c returns the face idx of the intersection, or -1 otherwise
            idx = (<FaceList>face_set).intersect_c(ray, point)
            if idx >= 0:
                nearest_set = j
                nearest_idx = idx
        if nearest_idx >= 0:
            #print "GET FACE", nearest.face_idx, len(all_faces)
            face = all_faces[nearest_idx]
            face.count += 1
            #print "ray length", ray.length
            point = addvv_(ray.origin, multvs_(ray.direction, ray.length))
            face_set = (<FaceList>(face_sets[nearest_set])) 
            orient = face_set.compute_orientation_c(face, point)
            #print "s normal", normal
            ### Clear the child_rays structure
            child_rays.n_rays = 0
            (<InterfaceMaterial>(face.material)).eval_child_ray_c(ray, i, 
                                                    point,
                                                    orient,
                                                    child_rays
                                                    )
            trace_parabasal_rays(gausslet, child_rays, face, face_set, new_gausslets, max_length)
            
    for j in range(n_decomp):
        face = decomp_faces[j]
        if face.count:
            (<InterfaceMaterial>(face.material)).eval_decomposed_rays_c(new_gausslets)
            face.count = 0
            
    new_gausslets.reset_length_c(max_length)
    return new_gausslets


cdef void trace_parabasal_rays(gausslet_t *g_in, RayCollection base_rays, Face face, FaceList face_set, 
                          GaussletCollection new_gausslets, double max_length):
    cdef:
        unsigned int i,j
        para_t *para_ray
        gausslet_t gausslet
        vector_t ray_end, point[6]
        orientation_t orient[6]
        InterfaceMaterial material = (<InterfaceMaterial>(face.material))
        
    for j in range(6):
        para_ray = g_in.para + j
        ray_end = addvv_(para_ray.origin, 
                        multvs_(para_ray.direction, 
                                max_length))
        if face_set.intersect_para_c(para_ray, ray_end, face):
            return
        point[j] = addvv_(para_ray.origin, multvs_(para_ray.direction, para_ray.length))
        orient[j] = face_set.compute_orientation_c(face, point[j])
    
    for i in range(base_rays.n_rays):
        gausslet.base_ray = base_rays.rays[i]
        for j in range(6):
            para_ray = g_in.para + j            
            gausslet.para[j] = material.eval_parabasal_ray_c(base_rays.rays + i,
                                                            para_ray.direction, #incoming ray direction
                                                            point[j], #position of intercept
                                                            orient[j],
                                                            gausslet.base_ray.ray_type_id, #indicates if it's a transmitted or reflected ray 
                                                            )
        new_gausslets.add_gausslet_c(gausslet)
            
            

def transform(Transform t, p):
    cdef vector_t p1, p2
    assert isinstance(t, Transform)
    assert len(p)==3
    p1.x = p[0]
    p1.y = p[1]
    p1.z = p[2]
    p2 = transform_c(t.trans, p1)
    return (p2.x, p2.y, p2.z)
    
def get_ray_size():
    return sizeof( ray_t )



from raypier.tracer import RayTraceModel
from raypier.parabolics import OffAxisParabloid
from raypier.sources import ConfocalRaySource

h = 25.4

oap = OffAxisParabloid(diameter=50.4,
                       height = h,
                       EFL = 50.4)

source = ConfocalRaySource(focus = (0,0,0),
                           direction = (1,0,0),
                           working_dist= 10.0,
                           number=8,
                           rings=2,
                           theta=15.0,
                           export_pipes=True)

model = RayTraceModel(optics=[oap],
                      sources=[source])

model.configure_traits()


#!/usr/bin/env python

#import pyximport
#pyximport.install()

import sys
# sys.path.append('..')
from raypier import cfaces, ctracer
import unittest
from math import sqrt


class AnOwner(object):
    def __init__(self, **kwds):
        self.__dict__.update(kwds)

class TestCircularFace(unittest.TestCase):
    def test_params(self):
        c = cfaces.CircularFace()
        print(c.params)
        self.assertEqual(c.params, ['diameter', 'offset'])

    def test_update(self):
        o = AnOwner(diameter=5.5, offset=6.6)
        c = cfaces.CircularFace(owner=o)
        c.update()
        self.assertEqual((c.diameter, c.offset), (o.diameter, o.offset))

    def test_intersection(self):
        o = AnOwner(diameter=5.5, offset=6.6)
        c = cfaces.CircularFace(owner=o)
        dist = c.intersect((0,0,-2), (0,0,2))
        self.assertEqual(dist, 2.0)

    def test_intersection2(self):
        o = AnOwner(diameter=5.5, offset=6.6)
        c = cfaces.CircularFace(owner=o)
        dist = c.intersect((-1,0,-1), (1,0,1))
        self.assertEqual(dist, sqrt(2.0))

    def test_face_set_intersection(self):
        o = AnOwner(diameter=5.5, offset=6.6)
        c = cfaces.CircularFace(owner=o)
        c.idx = 7
        fl = ctracer.FaceList()
        fl.faces = [c]
        r = ctracer.Ray(origin=(-1,0,-1), direction=(1,0,1), length=10)
        idx = fl.intersect(r, 20)
        self.assertEqual(r.end_face_idx, 7)
        self.assertEqual(r.length, sqrt(2.0))

    def test_miss(self):
        o = AnOwner(diameter=5.5, offset=6.6)
        c = cfaces.CircularFace(owner=o)
        dist = c.intersect((6,0,-2),(6,0,2))
        import numpy
        self.assertEqual(dist, 0.0)


class TestExtrudedFace(unittest.TestCase):
    def setUp(self):
        o = AnOwner()
        self.f = cfaces.ExtrudedPlanarFace(owner=o, z1=-1, z2=3,
                                            x1=-2.0,y1=2.0, x2=2.0, y2=-2.0)

    def test_intersection_1(self):
        dist = self.f.intersect((-5,0,0),(5,0,0))
        self.assertEqual(dist, 5.0)

    def test_miss_1(self):
        dist = self.f.intersect((-5,0,4),(5,0,4))
        self.assertEqual(dist, 0.0)

    def test_miss_2(self):
        dist = self.f.intersect((-5,0,-2),(5,0,-0.1))
        self.assertEqual(dist, 0.0)

    def test_miss_3(self):
        dist = self.f.intersect((-5,2.1,-1),(5,2.1,1))
        self.assertEqual(dist, 0.0)

if __name__=="__main__":
    unittest.main()



"""routines used in the evaluation of the point-spread-function probes"""
from traits.api import HasTraits, Instance, Property, Float

from raypier.bases import normaliseVector, dotprod
from raypier.rays import collectRays

from scipy.interpolate import interp2d
from scikits.delaunay import NNInterpolator, Triangulation
import numpy

class PSF(HasTraits):
    owner = Instance(klass="raypier.bases.Probe")
    
    sample_spacing = Property(Float, depends_on="owner")
    
    def do_trace(self, input_rays, face_sequence):
        rays = input_rays
        for faceList in face_sequence:
            if len(faceList) > 1:
                print("rays more than 1")
                rayList = [f.trace_rays(rays) for f in faceList]
                rays = collectRays(*rayList)
            else:
                rays = faceList[0].trace_rays(rays)
                
        self.project_onto_exit_pupil(rays)
        return rays
    
    def project_onto_exit_pupil(self, rays):
        owner = self.owner
        #mask = rays.length != numpy.Infinity
        directions = normaliseVector(rays.direction)#[mask]
        E1 = rays.E1_amp
        E2 = rays.E2_amp
        amp = numpy.sqrt(E1*E1.conjugate() + E2*E2.conjugate()).real
        mean_dir = normaliseVector((directions*amp).mean(axis=0))
               
        exit_pupil_offset = owner.exit_pupil_offset
        aperture = exit_pupil_offset /(owner.aperture*2.)
        pupil_centre = owner.position - exit_pupil_offset * mean_dir
        
        #project onto pupil plane
        delta = rays.origin - pupil_centre
        height = dotprod(delta, mean_dir)
        aspect = dotprod(directions, mean_dir)
        new_origin = rays.origin - (height/aspect) * directions
        
        rays.origin = new_origin
        rays.offset_length = height/aspect
        return
    
    def evaluate_scalar_amp(self, rays, target):
        newaxis = numpy.newaxis
        direction = rays.direction[newaxis,newaxis,...]
        origin = rays.origin[newaxis,newaxis,...]
        optical_path = rays.cum_length[newaxis,newaxis,:,0]
        target = target[:,:,newaxis,:]
        
        delta = target - origin
        z = dotprod(direction, delta)
        r_bar = delta - (z*direction)
        r = numpy.sqrt(dotprod(r_bar,r_bar))[...,0]
        z = z[...,0]
        
        wavelen = self.owner.wavelength / 1000.0 #convert to mm
        
        ray_areas = rays.ray_areas[newaxis,newaxis,:]
        w0 = numpy.sqrt(ray_areas/numpy.pi) #beam waists
        z_r = ray_areas / wavelen #Rayleigh length
        w = w0 * numpy.sqrt(1 + (z/z_r)**2)
        R = z * ( 1 + (z_r/z)**2 )
        
        q = 1./( 1/R - (1j*wavelen)/(numpy.pi * w *w) )
        
        ik = -2j*numpy.pi/wavelen
        
        E = numpy.exp(ik*optical_path + (2*ik*r**2)/q)
        
        E_total = E.sum(axis=2)
        return E_total
    
    def resample_exit_pupil(self):
        sample_spacing = self.sample_spacing #FIXME
        axes = numpy.array([[1,0,0],[0,1,0],[0,0,1]])
        proj = dotprod(axes,mean_dir).argmax()
        best_ax = axes[proj]
        axis1 = normaliseVector(numpy.cross(best_ax, mean_dir))
        axis2 = normaliseVector(numpy.cross(axis1, mean_dir))
        
        grid_size = int(aperture/sample_spacing)
        
        A,B = numpy.ogrid[-aperture:aperture:grid_size*1j,
                          -aperture:aperture:grid_size*1j]
        
        #convert points to UV coords
        delta = new_origin - pupil_centre
        U = dotprod(delta, axis1)
        V = dotprod(delta, axis2)
        
        cells = rays.cells
        for cell in cells:
            pass
    
    def sum_wavelets(rays):
        pass
            
    


#    Copyright 2009, Teraview Ltd., Bryan Cole
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from OCC import BRepPrimAPI, BRep, STEPControl, TopoDS, gp, \
        BRepBuilderAPI, BRepAlgoAPI, BRepOffsetAPI, Geom, TopAbs, TopExp,\
        XCAFApp, STEPCAFControl, TDocStd, TCollection,\
        XCAFDoc, Quantity, TopLoc
from itertools import tee
import itertools
import numpy

def pairs(itr):
    a,b = tee(itr)
    next(b)
    return zip(a,b)

def sign(val):
    return val/abs(val)

def MakeVertex(pt): 
    vt = BRepBuilderAPI.BRepBuilderAPI_MakeVertex(gp.gp_Pnt(*pt))
    #print "make vertex", pt, vt
    return vt

def MakeEdge(v1, v2):
    e = BRepBuilderAPI.BRepBuilderAPI_MakeEdge(v1.Vertex(), v2.Vertex())
    #print "make edge", v1, v2, e
    return e

def toshape(builder):
    shape = builder.Shape()
    shape._builder = builder
    return shape

def position_shape(shape, centre, direction, x_axis):
    ax = gp.gp_Ax3()
    ax.SetLocation(gp.gp_Pnt(*centre))
    ax.SetDirection(gp.gp_Dir(*direction))
    ax.SetXDirection(gp.gp_Dir(*x_axis))
    
    tr = gp.gp_Trsf()
    tr.SetTransformation(ax, gp.gp_Ax3())
    
    loc = TopLoc.TopLoc_Location(tr)
    
    shape.Location(loc)
    #trans = BRepBuilderAPI.BRepBuilderAPI_Transform(shape, tr)
    #return toshape(trans)
    return shape

def make_box(position, direction, x_axis, dx, dy, dz):
    box = BRepPrimAPI.BRepPrimAPI_MakeBox(gp.gp_Pnt(-dx/2., -dy/2., 0.0),
                                          dx, dy, -dz)
    ax = gp.gp_Ax2(gp.gp_Pnt(*position), 
                   gp.gp_Dir(*direction),
                   gp.gp_Dir(*x_axis))
    ax3 = gp.gp_Ax3()
    trans = gp.gp_Trsf()
    trans.SetTransformation(gp.gp_Ax3(ax), ax3)
    t_box = BRepBuilderAPI.BRepBuilderAPI_Transform(box.Shape(), trans)
    return toshape(t_box)

def make_cylinder(position, direction, radius, length, offset, x_axis):
    cyl_ax = gp.gp_Ax2(gp.gp_Pnt(offset,0,0), 
                       gp.gp_Dir(0,0,1), 
                       gp.gp_Dir(1,0,0))
    cyl = BRepPrimAPI.BRepPrimAPI_MakeCylinder(cyl_ax, radius, length)
    ax = gp.gp_Ax2(gp.gp_Pnt(*position), 
                   gp.gp_Dir(*direction),
                   gp.gp_Dir(*x_axis))
    ax3 = gp.gp_Ax3()
    trans = gp.gp_Trsf()
    trans.SetTransformation(gp.gp_Ax3(ax), ax3)
    t_cyl = BRepBuilderAPI.BRepBuilderAPI_Transform(cyl.Shape(), trans)
    print(position, direction, radius, length)
    return toshape(t_cyl)

def make_cylinder_2(start, end, radius):
    start = numpy.asarray(start)
    end = numpy.asarray(end)
    direction = end-start
    length = numpy.sqrt((direction**2).sum())
    cyl = BRepPrimAPI.BRepPrimAPI_MakeCylinder(radius, length)
    ax = gp.gp_Ax3(gp.gp_Pnt(*start), gp.gp_Dir(*direction))
    trans = gp.gp_Trsf()
    trans.SetTransformation(ax, gp.gp_Ax3())
    t_cyl = BRepBuilderAPI.BRepBuilderAPI_Transform(cyl.Shape(), trans)
    return toshape(t_cyl)

def make_sphere(position, radius):
    sph = BRepPrimAPI.BRepPrimAPI_MakeSphere(radius)
    trans = gp.gp_Trsf()
    trans.SetTranslation(gp.gp_Vec(*position))
    t_sph = BRepBuilderAPI.BRepBuilderAPI_Transform(sph.Shape(), trans)
    return toshape(t_sph)

# def make_spherical_lens(position, direction, x_axis, 
#                         diameter, curvature1, curvature2, 
#                         centre_height1, centre_height2, offset):
#     radius = diameter/2.
#     lower_z = min(centre_height1+curvature1, centre_height2+curvature2,
#                   centre_height1, centre_height2)
#     upper_z = max(centre_height1+curvature1, centre_height2+curvature2,
#                   centre_height1, centre_height2)
#     length = upper_z-lower_z
#     cyl_ax = gp.gp_Ax2(gp.gp_Pnt(offset,0,lower_z), 
#                        gp.gp_Dir(0,0,1), 
#                        gp.gp_Dir(1,0,0))
#     cyl = BRepPrimAPI.BRepPrimAPI_MakeCylinder(cyl_ax, radius, length)
#     
#     
    

def scale(sx, sy, sz):
    t = gp.gp_GTrsf()
    t.SetValue(1,1,sx)
    t.SetValue(2,2,sy)
    t.SetValue(3,3,sz)
    return t

def make_ellipsoid(focus1, focus2, major_axis):
    """
    @param focus1: length 3 sequence giving first focus location
    @param focus2: length 3 sequence giving second focus location
    @param path_length: major axis length
    """
    f1 = numpy.asarray(focus1)
    f2 = numpy.asarray(focus2)
    direction = -(f1 - f2)
    centre = (f1 + f2)/2.
    sep = numpy.sqrt((direction**2).sum())
    minor_axis = numpy.sqrt( major_axis**2 - (sep/2.)**2 )
    
    sphere = BRepPrimAPI.BRepPrimAPI_MakeSphere(minor_axis)
    
    scale = gp.gp_GTrsf()
    scale.SetValue(3,3, major_axis/minor_axis)
    
    ellipse = BRepBuilderAPI.BRepBuilderAPI_GTransform(sphere.Shape(), scale)
    
    loc = gp.gp_Ax3()
    loc.SetLocation(gp.gp_Pnt(*centre))
    loc.SetDirection(gp.gp_Dir(*direction))
    
    tr = gp.gp_Trsf()
    tr.SetTransformation(loc, gp.gp_Ax3())
    
    trans = BRepBuilderAPI.BRepBuilderAPI_Transform(ellipse.Shape(), tr)
    shape = toshape(trans)
    return shape

def make_ellipsoid_2(focus1, focus2, major_axis):
    f1 = numpy.asarray(focus1)
    f2 = numpy.asarray(focus2)
    direction = -(f1 - f2)
    centre = (f1 + f2)/2.
    sep = numpy.sqrt((direction**2).sum())
    minor_axis = numpy.sqrt( major_axis**2 - (sep/2.)**2 )
    
    el = gp.gp_Elips(gp.gp_Ax2(gp.gp_Pnt(0,0,0), gp.gp_Dir(1,0,0)), 
                          major_axis, minor_axis)
    edge1 = BRepBuilderAPI.BRepBuilderAPI_MakeEdge(el, 
                                                  gp.gp_Pnt(0,0,major_axis),
                                                  gp.gp_Pnt(0,0,-major_axis))
    edge2 = BRepBuilderAPI.BRepBuilderAPI_MakeEdge(gp.gp_Pnt(0,0,-major_axis),
                                                  gp.gp_Pnt(0,0,major_axis))
    
    wire = BRepBuilderAPI.BRepBuilderAPI_MakeWire()
    wire.Add(edge1.Edge())
    wire.Add(edge2.Edge())
    
    face = BRepBuilderAPI.BRepBuilderAPI_MakeFace(wire.Wire())
    
    el = BRepPrimAPI.BRepPrimAPI_MakeRevol(face.Shape(), 
                                           gp.gp_Ax1(gp.gp_Pnt(0,0,0),
                                                     gp.gp_Dir(0,0,1)),
                                           numpy.pi*2)
    
    loc = gp.gp_Ax3()
    loc.SetLocation(gp.gp_Pnt(*centre))
    loc.SetDirection(gp.gp_Dir(*direction))
    
    tr = gp.gp_Trsf()
    tr.SetTransformation(loc, gp.gp_Ax3())
    
    trans = BRepBuilderAPI.BRepBuilderAPI_Transform(el.Shape(), tr)
    shape = toshape(trans)
    return shape

def make_ellipsoid_mirror(f1, f2, major_axis, 
                          x_bounds, y_bounds, z_bounds,
                          centre, direction, x_axis):
    ellipsoid = make_ellipsoid(f1, f2, major_axis)
    P1 = gp.gp_Pnt(x_bounds[0], y_bounds[0], z_bounds[0])
    P2 = gp.gp_Pnt(x_bounds[1], y_bounds[1], z_bounds[1])
    block = BRepPrimAPI.BRepPrimAPI_MakeBox(P1, P2)
    
    ##comment these out to reinstate the cut
    #t_block = position_shape(toshape(block), centre, direction, x_axis)
    #t_ellipse = position_shape(ellipsoid, centre, direction, x_axis)
    #return make_compound([toshape(block), ellipsoid])
    
    #cut = toshape(BRepAlgoAPI.BRepAlgoAPI_Cut(toshape(block), ellipsoid))
    cut = make_compound([toshape(block), ellipsoid])
    return position_shape(cut, centre, direction, x_axis)

def make_interp_parabola(FL, rmin, rmax, segments=50):
    A = 1./(4*FL)
    x = numpy.linspace(rmin, rmax, segments)
    y = (A * x**2) - FL
    
    points = [(X,0,Z) for X,Z in zip(x,y)]
    points.append((x[0],0,y[-1]))
    
    def pairs(itr):
        a,b = itertools.tee(itr)
        next(b)
        return zip(a,b)
    
    edges = (BRepBuilderAPI.BRepBuilderAPI_MakeEdge(
                    gp.gp_Pnt(*p1), gp.gp_Pnt(*p2))
                    for p1, p2 in pairs(points))
    last_edge = BRepBuilderAPI.BRepBuilderAPI_MakeEdge(
                    gp.gp_Pnt(*points[-1]), 
                    gp.gp_Pnt(*points[0]))
    
    wire = BRepBuilderAPI.BRepBuilderAPI_MakeWire()
    for e in edges:
        wire.Add(e.Edge())
    wire.Add(last_edge.Edge())
    
    face = BRepBuilderAPI.BRepBuilderAPI_MakeFace(wire.Wire())
    
    ax = gp.gp_Ax1(gp.gp_Pnt(0,0,0),
                   gp.gp_Dir(0,0,1))
    
    revol = BRepPrimAPI.BRepPrimAPI_MakeRevol(face.Shape(), ax)
    
    return revol.Shape()

def make_true_para(FL, rmin, rmax):
    """
    makes a parabloid, with rotation axis along Z and focus
    at the origin
    """
    ax = gp.gp_Ax2(gp.gp_Pnt(0.,0.,-FL), #origin
                   gp.gp_Dir(1.,0.,0.), #main direction is Z
                   gp.gp_Dir(0.,0.,1.)) #X Direction is X
    para = Geom.Geom_Parabola(ax, FL)
    h_para = Geom.Handle_Geom_Parabola(para)
    
    ax2 = gp.gp_Ax2(gp.gp_Pnt(0,0,0), #origin
                   gp.gp_Dir(0.,0.,1.), #main direction is Z
                   gp.gp_Dir(1.,0.,0.)) #X Direction is X
    
    pbl_shape = BRepPrimAPI.BRepPrimAPI_MakeRevolution(ax2, h_para,
                                                       rmin, rmax)
    return pbl_shape.Shape()

def make_OAP(EFL, diameter, height, centre, direction, x_axis):
    FL = EFL/2. #focal length
    radius = diameter/2.
    outside = EFL + radius
    inside = EFL - radius
    length = (outside**2)/(4.*FL) - FL + height
    
    #pbl_shape = make_true_para(FL, 5.0, outside+1)
    pbl_shape = make_interp_parabola(FL, inside-1, outside+1)
    
    ax3 = gp.gp_Ax2(gp.gp_Pnt(EFL,0,-height), #origin
                   gp.gp_Dir(0.,0.,1.), #main direction is X
                   gp.gp_Dir(0.,1.,0.)) #X Direction is Y
    cyl_solid = BRepPrimAPI.BRepPrimAPI_MakeCylinder(ax3, radius, length)
    
    
    cut = BRepAlgoAPI.BRepAlgoAPI_Cut(cyl_solid.Shape(), 
                                      pbl_shape)
        
    loc= position_shape(toshape(cut), centre, 
                          direction, x_axis)
    #nurbs = BRepBuilderAPI.BRepBuilderAPI_NurbsConvert(loc)
    return loc #toshape(nurbs)

def make_spherical_lens(CT, diameter, curvature, 
                        centre, direction, x_axis):
    cax = gp.gp_Ax2(gp.gp_Pnt(0,0,CT-curvature),
                    gp.gp_Dir(0,1,0),
                    gp.gp_Dir(1,0,0))
    circ = Geom.Geom_Circle(cax, curvature)
    h_circ = Geom.Handle_Geom_Circle(circ)
    
    r = diameter/2.
    h2 = CT - curvature + numpy.sqrt(curvature**2 - r**2)
    p1 = gp.gp_Pnt(0,0,CT)
    p2 = gp.gp_Pnt(r,0,h2)
    p3 = gp.gp_Pnt(r,0,0)
    p4 = gp.gp_Pnt(0,0,0)
    
    #ps = p1,p2,p3,p4
    #vs = [BRepBuilderAPI.BRepBuilderAPI_MakeVertex(p) for p in ps]
    
    e1 = BRepBuilderAPI.BRepBuilderAPI_MakeEdge(h_circ, p1,
                                                p2)    
    e2 = BRepBuilderAPI.BRepBuilderAPI_MakeEdge(p2,p3)
    e3 = BRepBuilderAPI.BRepBuilderAPI_MakeEdge(p3,p4)
    e4 = BRepBuilderAPI.BRepBuilderAPI_MakeEdge(p4,p1)
    
    wire = BRepBuilderAPI.BRepBuilderAPI_MakeWire()
    for e in (e1,e2,e3,e4):
        print(e)
        wire.Add(e.Edge())
    
    face = BRepBuilderAPI.BRepBuilderAPI_MakeFace(wire.Wire())
    
    ax = gp.gp_Ax1(gp.gp_Pnt(0,0,0),
                   gp.gp_Dir(0,0,1))
    solid = BRepPrimAPI.BRepPrimAPI_MakeRevol(face.Shape(), ax)
    
    return position_shape(toshape(solid), centre, direction, x_axis)


def make_spherical_lens2(CT1, CT2, diameter, 
                         curvature1, curvature2, 
                        centre, direction, x_axis):
    cax = gp.gp_Ax2(gp.gp_Pnt(0,0,CT1-curvature1),
                    gp.gp_Dir(0,sign(curvature1),0),
                    gp.gp_Dir(1,0,0))
    circ = Geom.Geom_Circle(cax, abs(curvature1))
    h_circ = Geom.Handle_Geom_Circle(circ)
    
    cax2 = gp.gp_Ax2(gp.gp_Pnt(0,0,CT2-curvature2),
                    gp.gp_Dir(0,-sign(curvature2),0),
                    gp.gp_Dir(1,0,0))
    circ2 = Geom.Geom_Circle(cax2, abs(curvature2))
    h_circ2 = Geom.Handle_Geom_Circle(circ2)
    
    r = diameter/2.
    
    h2 = CT1 - curvature1 + numpy.sqrt(curvature1**2 - r**2)*sign(curvature1)
    h3 = CT2 - curvature2 + numpy.sqrt(curvature2**2 - r**2)*sign(curvature2)
    p1 = gp.gp_Pnt(0,0,CT1)
    p2 = gp.gp_Pnt(r,0,h2)
    p3 = gp.gp_Pnt(r,0,h3)
    p4 = gp.gp_Pnt(0,0,CT2)
    
    e1 = BRepBuilderAPI.BRepBuilderAPI_MakeEdge(h_circ, p1, p2)    
    e2 = BRepBuilderAPI.BRepBuilderAPI_MakeEdge(p2,p3)
    e3 = BRepBuilderAPI.BRepBuilderAPI_MakeEdge(h_circ2, p3,p4)
    e4 = BRepBuilderAPI.BRepBuilderAPI_MakeEdge(p4,p1)
    
    wire = BRepBuilderAPI.BRepBuilderAPI_MakeWire()
    for e in (e1,e2,e3,e4):
        print(e)
        wire.Add(e.Edge())
    
    face = BRepBuilderAPI.BRepBuilderAPI_MakeFace(wire.Wire())
    
    ax = gp.gp_Ax1(gp.gp_Pnt(0,0,0),
                   gp.gp_Dir(0,0,1))
    solid = BRepPrimAPI.BRepPrimAPI_MakeRevol(face.Shape(), ax)
    
    return position_shape(toshape(solid), centre, direction, x_axis)
    

def make_wire(listOfPoints, close=False):
    vertices = [BRepBuilderAPI.BRepBuilderAPI_MakeVertex(gp.gp_Pnt(*p))
                for p in listOfPoints]
    edges = [BRepBuilderAPI.BRepBuilderAPI_MakeEdge(v1.Vertex(),v2.Vertex())
         for v1,v2 in pairs(vertices)]
    if close:
        edges.append(BRepBuilderAPI.BRepBuilderAPI_MakeEdge(vertices[-1].Vertex(),
                                                            vertices[0].Vertex()))
    wire = BRepBuilderAPI.BRepBuilderAPI_MakeWire()
    for e in edges:
        wire.Add(e.Edge())
    return toshape(wire)

def make_compound(listOfShapes):
    aRes = TopoDS.TopoDS_Compound()
    aBuilder = BRep.BRep_Builder()
    aBuilder.MakeCompound(aRes)
    for item in listOfShapes:
        aBuilder.Add(aRes, item)
    aBuilder._InputShapes = listOfShapes
    aRes._builder = aBuilder
    return aRes    

def make_rays_both(listOfRays, scale):
    wires = make_rays_wires(listOfRays, scale)
    pipes = make_rays_pipes(listOfRays, scale)
    return make_compound([wires, pipes])

def make_rays_wires(listOfRays, scale=None):
    raysItr = iter(listOfRays)
    first = next(raysItr)
    def MakeVertex(pt): 
        vt = BRepBuilderAPI.BRepBuilderAPI_MakeVertex(gp.gp_Pnt(*pt))
        #print "make vertex", pt, vt
        return vt
    def MakeEdge(v1, v2):
        e = BRepBuilderAPI.BRepBuilderAPI_MakeEdge(v1.Vertex(), v2.Vertex())
        #print "make edge", v1, v2, e
        return e
    wires = [BRepBuilderAPI.BRepBuilderAPI_MakeWire() 
             for i in range(first.origin.shape[0])]
    v_start = [MakeVertex(pt) for pt in first.origin]
    v_end = [MakeVertex(pt) for pt in first.termination]
    first_edges = [MakeEdge(v1,v2) for v1, v2 in zip(v_start, v_end)]
    for edge, wire in zip(first_edges, wires):
        wire.Add(edge.Edge())
    id_map = list(range(len(wires)))
    for rays in raysItr:
        id_map = [id_map[pid] for pid in rays.parent_idx]
        v_start = [v_end[pid] for pid in rays.parent_idx]
        v_end = [MakeVertex(pt) for pt in rays.termination]
        edges = [MakeEdge(v1,v2) for v1, v2 in zip(v_start, v_end)]
        for edge, w_id in zip(edges, id_map):
            wires[w_id].Add(edge.Edge())
    return make_compound([w.Shape() for w in wires])

def make_rays_pipes(listOfRays, radius=0.1):
    itrRays = iter(listOfRays)
    first = next(itrRays)
    v_lists = [[s,e] for s,e in zip(first.origin, first.termination)]
    id_map = list(range(len(v_lists)))
    for rays in itrRays:
        id_map = [id_map[pid] for pid in rays.parent_idx]
        v_end = list(rays.termination)
        for w_id, v in zip(id_map, v_end):
            v_lists[w_id].append(v)
        
    ### Fusing shape is finicky
    #_fuse = fuse_shapes
    _fuse = make_compound
    shapes=[]    
    for vlist in v_lists:
        cylinders = (make_cylinder_2(s, e, radius) for s,e in pairs(vlist))
        spheres = (make_sphere(p, radius) for p in vlist[1:-1])
        interleave = list(next(it) for it in itertools.cycle([cylinders,
                                                          spheres]))
        shapes.append(_fuse(interleave))
        #shapes.extend(cylinders)
    return make_compound(shapes)

def fuse_shapes(shapeList):
    s1 = shapeList[0]
    _builders = []
    for s2 in shapeList[1:]:
        fuse = BRepAlgoAPI.BRepAlgoAPI_Fuse(s1, s2)
        _builders.append(fuse)
        s1 = fuse.Shape()
    s1._builders = _builders
    return s1

def export_shapes(shapeList, filename):
    print(shapeList)
    step_export = STEPControl.STEPControl_Writer()
    for shape in shapeList:
        step_export.Transfer(shape,STEPControl.STEPControl_AsIs)
    step_export.Write(str(filename))
    
def get_color_map():
    c = [a for a in dir(Quantity) if a.startswith("Quantity_NOC_")]
    names = [a.split('_')[-1].lower() for a in c]
    vals = [getattr(Quantity, n) for n in c]
    return dict(list(zip(names,vals)))
    
def export_shapes2(shapeList, filename, colorList=[]):
    h_doc = TDocStd.Handle_TDocStd_Document()
    app = XCAFApp.XCAFApp_Application_GetApplication().GetObject()
    app.NewDocument(TCollection.TCollection_ExtendedString("MDTV-CAF"),h_doc)
    doc = h_doc.GetObject()
    h_shape_tool = XCAFDoc.XCAFDoc_DocumentTool().ShapeTool(doc.Main())
    h_color_tool = XCAFDoc.XCAFDoc_DocumentTool().ColorTool(doc.Main())
    shape_tool = h_shape_tool.GetObject()
    color_tool = h_color_tool.GetObject()
    top_label = shape_tool.NewShape()
    
    colorList.extend(["brown"]*(len(shapeList) - len(colorList)))
    cmap = get_color_map()
    
    tr = gp.gp_Trsf()
    loc = TopLoc.TopLoc_Location(tr)
    print(colorList)
    print(shapeList)
    colorMap = dict((c, Quantity.Quantity_Color(cmap.get(c, 133))) for c in colorList)
    
    for shape, color in zip(shapeList, colorList):
        if not shape:
            continue
        print("color:", color, "shape", shape)
        label = shape_tool.AddShape(shape, False, True)
        ref_label = shape_tool.AddComponent(top_label, label, loc)
        c = colorMap[color]
        color_tool.SetColor(ref_label, c, XCAFDoc.XCAFDoc_ColorGen)
        
    mode = STEPControl.STEPControl_AsIs
    writer = STEPCAFControl.STEPCAFControl_Writer()
    writer.Transfer(h_doc, mode)
    writer.Write(str(filename))
    
if __name__=="__main__":
    cyl = make_cylinder((1,2,3),(7,6,5),5,2, 0, (1,1,1) )
    export_shapes2([cyl]*5, "test_step_export.stp")


from raypier.lenses import PlanoConvexLens
from raypier.sources import GaussianBeamRaySource
from raypier.tracer import RayTraceModel

m1 = PlanoConvexLens(name="L1",
                centre=(0,-45, 0),
                direction=(0.,-1,0), #(0,y_dir,-small_height),
                diameter=10.,
                CT=4.,
                curvature=15.0,
                n_inside=1.6)

source = GaussianBeamRaySource(origin=(0,0,0),
                           direction=(0,-1,0),
                           beam_waist=1000.0, #in microns
                           number=20,
                           wavelength=100)

model = RayTraceModel(sources=[source],
                      optics=[m1])

model.configure_traits()
#model.update = True
#print source.traced_rays


"""
Traits wrappers for Distortion objects
"""

from traits.api import Instance, Float, HasTraits, observe, Event, List, Tuple, Int
from traitsui.api import View, VGroup, Item
from traitsui.editors import ListEditor, TupleEditor

from .editors import NumEditor
from .core import cdistortions, ctracer
from traits.observation.events import ListChangeEvent
from traits.observation._trait_change_event import TraitChangeEvent


class BaseDistortion(HasTraits):
    updated = Event()
    c_distortion = Instance(ctracer.Distortion, ())
    
    
class NullDistortion(BaseDistortion):
    """Any zero-amplitude distortion. Mainly used in testing.
    """
    pass


class SimpleTestZernikeJ7(BaseDistortion):
    """
    An implementation of a specific Zernike polynomial 
    distortion, using J=7. Mainly used in testing.
    """
    
    #: Defines the unit-radius for the Zernike polynomial function.
    unit_radius = Float(10.0)
    
    #: RMS amplitude of the distortion
    amplitude = Float(0.001)
    
    #: The underlying core.Distortion object 
    c_distortion = Instance(cdistortions.SimpleTestZernikeJ7, ())
    
    traits_view = View(VGroup(
                    Item("unit_radius", editor=NumEditor),
                    Item("amplitude", editor=NumEditor)
                        )
                    )
    
    def _c_distortion_default(self):
        return cdistortions.SimpleTestZernikeJ7(unit_radius=self.unit_radius,
                                                amplitude=self.amplitude)
        
    @observe("unit_radius, amplitude")
    def on_params_change(self, evt):
        d = self.c_distortion
        setattr(d, evt.name, evt.new)
        self.updated = True
        
        
coef_editor = TupleEditor(cols=2, labels=["j", "value"])
        
        
class ZernikeSeries(BaseDistortion):
    """
    Represents a general surface distortion in terms of a sequence of Zernike polynomials.
    The Zernike terms are identified using the ANSI standard single-index notation (J).
    
    The standard normalisation is used such that the amplitude coefficients give the RMS
    deviation of the surface over the unit-disk.
    
    A recursive algorithm is used for evaluation of the function with caching
    for efficient evaluation where many non-zero coefficients exist.
    
    Both the given parameters are traits on this object and updates to either 
    one will automatically update the internal state of the object (and trigger
    re-tracing of the model).
    
    This class is intended to be used with the DistortionSurface to wrap any underlying ShapedFace.
    """
    
    #: The unit-radius for the Zernike polynomial radial function
    unit_radius = Float(10.0)
    
    #: A list of (int,float) tuples giving the amplitudes of all non-zero Zernike coefficients.
    #:     in terms of the ANSI standard single index J = (n*(n+1) + m)/2.
    #:     Any number of coefficients can be given.
    coefficients = List(Tuple(Int,Float))
    
    c_distortion = Instance(cdistortions.ZernikeDistortion)
    
    traits_view = View(VGroup(
                    Item("unit_radius", editor=NumEditor),
                    Item("coefficients", editor=ListEditor(editor=coef_editor))
                        )
                    )
    
    def __init__(self, *args, **kwds):
        coefs = dict(kwds.get("coefficients", []))
        if args:
            coefs.update(args[0])
        jks = {k for k in kwds if k.startswith("j")}
        coefs.update({int(k[1:]):float(kwds[k]) for k in jks})
        for k in jks:
            del kwds[k]
        kwds['coefficients'] = sorted(coefs.items())
        super().__init__(**kwds)
    
    def _c_distortion_default(self):
        return cdistortions.ZernikeDistortion(self.coefficients,
                                              unit_radius=self.unit_radius)
        
    @observe("coefficients.items, unit_radius")
    def on_params_change(self, evt):
        if isinstance(evt, TraitChangeEvent) and evt.name == "unit_radius":
            self.c_distortion.unit_radius = evt.new
        else:
            self.c_distortion.set_coefs(self.coefficients)
        self.updated = True
        


import pyximport
pyximport.install(language_level=3)

from raypier.tracer import RayTraceModel
from raypier.lenses import PlanoConicLens, PlanoConvexLens, BiConicLens, AsphericLens
from raypier.sources import ConfocalRaySource

lens = PlanoConicLens(centre=(20,0,0),
                       direction=(1,0,0),
                       curvature=25.0,
                       diameter=25.4,
                       conic_const = 0.0,
                       CT=6.0)
# 
# lens = AsphericLens(centre=(20,0,0),
#                     direction=(1,0,0),
#                     CT=3.8*5,
#                     diameter=4.7*5,
#                     A_curvature = -5*3.200655,
#                     A_conic = -4.321649,
#                     A4 = -5.521153E-3/(5**3),
#                     A6 = 1.981378E-3/(5**5),
#                     A8 = -4.782553E-4/(5**7),
#                     A10 = 7.328134E-5/(5**9),
#                     B_curvature = 5*3.200655,
#                     B_conic = -4.321649,
#                     B4 = 5.521153E-3/(5**3),
#                     B6 = -1.981378E-3/(5**5),
#                     B8 = 4.782553E-4/(5**7),
#                     B10 = -7.328134E-5/(5**9),
#                     )
# 
# lens = BiConicLens(centre=(20,0,0),
#                        direction=(1,0,0),
#                        curvature=25.0,
#                        curvature2=-25.0,
#                        diameter=25.4,
#                        conic_const = 0.0,
#                        conic_const2 = 0.0,
#                        CT=12.0)

# lens2 = PlanoConvexLens(centre=(20,0,0),
#                        direction=(1,0,0),
#                        curvature=25.0,
#                        diameter=25.4,
#                        conic_const = 0.0,
#                        CT=6.0)

#print("params", lens.params)
for f in lens.faces.faces:
    print(f.params)
 
source = ConfocalRaySource(focus = (0,0,0),
                           direction = (1,0,0),
                           working_dist= 100.0,
                           theta=15.0)
 
model = RayTraceModel(optics=[lens],
                      sources=[source])

model.configure_traits()



from raypier.splines import Extruded_interpolant
from raypier.sources import RectRaySource
from raypier.tracer import RayTraceModel
from raypier.core.cmaterials import OpaqueMaterial, PECMaterial, DielectricMaterial
from raypier.core.cfaces import ExtrudedPlanarFace
from raypier.dielectrictroughs import LDLF

import numpy as np

t = np.linspace(0, 2*np.pi, 100)
# dont use form like this:
#x = np.array([1,2,3,4,5,6,7,8,9,10])
# because these are used as integers and decimals are not used!
# range() also creates a list of integers, so use numpy.linspace
#y = -x*x+30*30*x

xs = 30*np.sin(t)
ys = 30*np.cos(t)

for s,x in enumerate(xs):
    if x<0:
        xs[s]=x/2. 
xs = xs+50
#give profile as list of x values then y values.  this profile is evaluated by splprep
#the smaller smoothness is, the more accurate the spline is, but the more faces it takes.
#try to make s as small as possible without making too many faces. (20 faces isn't unreasonable)

test = Extruded_interpolant(profile = [xs,ys], smoothness = 0, z_height_1 = -30, z_height_2=30, \
		material = DielectricMaterial(), n_inside=1.5, trace_ends= True,trace_top = False )

source = RectRaySource(origin=(50,0,0),
                            direction=(-1,0,0),
                            working_dist = 100.,
                            number=10,
                            length = 10,
			    width = 10,
			    randomness = True,
                            theta = 5.,
                            scale_factor=0.2)

'''
r1 = LDLF(material=PECMaterial(),
                z_height_1=-20.0,
                z_height_2=20.0,
                slat_width=20,
                ap_width=5,
                slant=75,
                n_inside=1.7)
'''


model = RayTraceModel(sources=[source],
                      optics=[test])
model.configure_traits()






from traits.api import Instance, HasStrictTraits, Property, cached_property,\
        Tuple, Float, Enum, Event, Str, observe, List, ComparisonMode
        
from traitsui.api import View, Item, VGroup, Label, Group, TupleEditor, ListEditor

from tvtk.api import tvtk

import numpy as np

from .core import cshapes
from .core.ctracer import Shape
from .editors import NumEditor


class BaseShape(HasStrictTraits):
    """Abstract base class for all Shape objects.
    """
    updated = Event()
    
    def __and__(self, other):
        return BooleanShape(self, other, operation="intersection")
    
    def __or__(self, other):
        return BooleanShape(self, other, operation="union")
    
    def __xor__(self, other):
        return BooleanShape(self, other, operation="difference")
    
    def __invert__(self):
        return InvertedShape(self)
    
    
    
    
class InvertedShape(BaseShape):
    """Performs a unary NOT operation on the given Shape
    """
    shape = Instance(BaseShape)
    cshape = Property(depends_on="shape")
    impl_func = Property(tvtk.ImplicitFunction, depends_on="shape")
    
    @cached_property
    def _get_cshape(self):
        return cshapes.InvertShape(self.shape)
        
    @cached_property
    def _get_impl_func(self):
        ### Can't figure out how to invert a VTK implicit function! crap.
        raise NotImplementedError
    

class BooleanShape(BaseShape):
    """
    Abstract base class for shape boolean operations.
    
    A BooleanShape owns two subshapes, `shape1` and `shape2`.
    These can be combined using any of the logical operations: difference,
    union, intersection and union-of-magnitudes.
    """
    ###Inputs
    shape1 = Instance(BaseShape)
    shape2 = Instance(BaseShape)
    operation_type = Enum("difference", "union", "intersection", "union_of_magnitudes")
    
    ###Outputs
    bounds = Property(depends_on="shape1.bounds, shape2.bounds")
    cshape = Property(depends_on="shape1, shape2, operation_type")
    impl_func = Property(depends_on="shape1, shape2, operation_type")
    cshape_type = Enum(cshapes.BooleanAND, cshapes.BooleanOR, cshapes.BooleanXOR, cshapes.InvertShape)
    
    traits_view = View(VGroup(
        Group(Item("shape1", style="custom", show_label=False, padding=-5),
              show_border=True,
              label="Shape 1")
              ,
        Item("operation_type", style="simple", show_label=False),
        Group(Item("shape2", style="custom", show_label=False, padding=-5),
              show_border=True,
              label="Shape 2"),
        padding = 0,
        show_border=True,
        label="Boolean Combination"
        ),
    )
    
    cshape_map = {"union": cshapes.BooleanOR, 
                      "difference": cshapes.BooleanXOR,
                      "union_of_magnitudes": cshapes.BooleanOR, 
                      "intersection": cshapes.BooleanAND}
    
    def __init__(self, shape1, shape2, **kwds):
        operation = kwds.pop("operation", "difference")
        super().__init__(**kwds)
        self.operation_type = operation
        self.shape1=shape1
        self.shape2=shape2
        
    @observe("shape1.updated, shape2.updated, operation_type")
    def _on_changed(self, evt):
        self.updated = True
        
    @cached_property
    def _get_bounds(self):
        b1 = self.shape1.bounds
        b2 = self.shape2.bounds
        return ( min(b1[0],b2[0]), max(b1[1],b2[1]), min(b1[2],b2[2]), max(b1[3],b2[3]) )
        
    @cached_property
    def _get_cshape(self):
        return self.cshape_map[self.operation_type](self.shape1.cshape,
                                  self.shape2.cshape)
        
    @cached_property
    def _get_impl_func(self):
        func = tvtk.ImplicitBoolean()
        func.add_function(self.shape1.impl_func)
        func.add_function(self.shape2.impl_func)
        func.operation_type = self.operation_type
        return func
        
        
class BasicShape(BaseShape):
    """
    Abstract base class for Shape primitives.
    """
    cshape = Instance(Shape)
    impl_func = Instance(tvtk.ImplicitFunction)
    
    bounds = Property()
    
    
    
class CircleShape(BasicShape):
    """
    A Shape representing a circle.
    
    Params
    ------
    
    centre - a 2-tuple giving the local centre coordinate
    
    radius - a float giving the circle radius
    
    """
    centre = Tuple(Float(0.0),Float(0.0))
    radius = Float(1.0)
    
    bounds = Property(depends_on="radius, centre")
    
    traits_view = View(
            Item("centre"),
            Item("radius", editor=NumEditor))
    
    @cached_property
    def _get_bounds(self):
        x,y = self.centre
        r = self.radius
        return (x-r,x+r,y-r,y+r)
    
    def _cshape_default(self):
        return cshapes.CircleShape(centre=self.centre, radius=self.radius)
    
    def _impl_func_default(self):
        x,y = self.centre
        func = tvtk.Cylinder(axis=(0,0,1), center=(x, y, 0),
                             radius=self.radius)
        return func
    
    @observe("radius, centre")
    def _update_imp_func(self, evt):
        x,y = self.centre
        func = self.impl_func
        func.radius = self.radius
        func.center = (x,y,0)
        self.cshape.radius = self.radius
        self.cshape.centre = (x,y)
        self.updated = True
        
        
class RectangleShape(BasicShape):
    """
    A Shape representing a rectangle.
    
    Params
    ------
    
    centre - a 2-tuple of floats giving the centre of the rectangle in local coordinates.
    
    width - the width of the rectangle
    
    height - the height of the rectangle
    """
    centre = Tuple(Float(0.0),Float(0.0))
    width = Float(5.0)
    height = Float(7.0)
    
    bounds = Property(depends_on="width, height, centre")
    
    _zextent = Float(10000.)
    
    traits_view = View(VGroup(Item("centre"),
                            Item("width", editor=NumEditor),
                              Item("height", editor=NumEditor)))
    
    @cached_property
    def _get_bounds(self):
        x,y = self.centre
        w=self.width/2.
        h = self.height/2.
        return (x-w,x+w,y-h,y+h)

    def _cshape_default(self):
        return cshapes.RectangleShape(centre=self.centre,
                                      width=self.width,
                                      height=self.height)
        
    def _impl_func_default(self):
        x,y = self.centre
        w=self.width/2.
        h = self.height/2.
        z = self._zextent
        func = tvtk.Box(bounds=(x-w,x+w,y-h,y+y,-z,z))
        return func
    
    @observe("width, height, centre")
    def _update_imp_func(self, evt):
        func = self.impl_func
        x,y = self.centre
        w=self.width/2.
        h = self.height/2.
        z = self._zextent
        func.bounds = (x-w,x+w,y-h,y+h,-z,z)
        cshape = self.cshape
        cshape.width = self.width
        cshape.height = self.height
        cshape.centre = (x,y)
        self.updated = True
        
        
coord_editor = TupleEditor(cols=2, labels=["x", "y"])
    
    
class PolygonShape(BasicShape):
    """
    A general polygon shape outline, defined in terms of a list of 2-tuples for the 
    (x,y) corner points in the local coordinate system of the optic.
    """
    
    #: A list of (x,y) coordinates. There should be at least 3 points in the list.
    #: The shape boundary should be non-self-intersecting.
    coordinates = List(trait=Tuple(Float,Float), comparison_mode=ComparisonMode.identity) #List()
    
    _coords_evt = Event()
    
    #: A property giving the bounding box for the shape.
    bounds = Property(depends_on="_coords_evt")
    
    traits_view = View(Item("coordinates", editor=ListEditor(editor=coord_editor)))
    
    
    def _cshape_default(self):
        return cshapes.PolygonShape(coordinates=np.asarray(self.coordinates))
    
    def _coordinates_default(self):
        return [(-1.,-1.),(0.,1.),(1.,-1.)]
    
    @cached_property
    def _get_bounds(self):
        coords = self.coordinates
        xmin = min(a[0] for a in coords)
        xmax = max(a[0] for a in coords)
        ymin = min(a[1] for a in coords)
        ymax = max(a[1] for a in coords)
        return (xmin,xmax,ymin,ymax)
    
    def _impl_func_default(self):
        points = [(a[0],a[1],0.0) for a in self.coordinates]
        func = tvtk.ImplicitSelectionLoop(loop=points,
                                          normal=(0.,0.,1.))
        return func
    
    @observe("coordinates.items")
    def _update_coords(self, evt):
        self._coords_evt = True 
        coords = self.coordinates
        points = [(a[0],a[1],0.0) for a in coords]
        self.impl_func.loop = points
        self.cshape.coordinates = np.array(coords)
        self.updated=True
        
        
class HexagonShape(PolygonShape):
    """A Hexagon shape, provided as a convenience. Subclasses PolygonShape
    """
    
    #: The centre of the hexagon
    centre = Tuple(Float(0.0),Float(0.0))
    
    #: The distance from the centre to the corners of the hexagon
    radius = Float(25.0)
    
    #: A rotation angle to be applied to the hexagon, in degrees
    rotation = Float(0.0)
    
    traits_view = View(VGroup(
            Item("centre", editor=TupleEditor(cols=2,labels=["x","y"])),
            Item("radius", editor=NumEditor),
            Item("rotation", editor=NumEditor)
            ))
    
    def make_coords(self):
        r = self.radius
        x,y = self.centre
        rot = self.rotation*np.pi/180.0
        angle = (rot + i*np.pi/3 for i in range(6))
        coords = [(r*np.cos(th)+x, r*np.sin(th)+y) for th in angle]
        return coords
    
    def _coordinates_default(self):
        return self.make_coords()
    
    @observe("radius, rotation, centre")
    def _radius_updated(self, evt):
        self.coordinates = self.make_coords()



from .bases import Traceable, NumEditor
from .core.cmaterials import CircularApertureMaterial, RectangularApertureMaterial
from .core.cfaces import CircularFace, RectangularFace
from .core.ctracer import FaceList

from traits.api import Float, on_trait_change, Instance, Property, Str, Bool, Constant
from traitsui.api import View, Item, VGroup
from tvtk.api import tvtk

import numpy
from scipy.special import erf


class BaseAperture(Traceable):
    edge_width = Float(2.0)
    invert = Bool(False)
    
    sc = Instance(tvtk.ProgrammableAttributeDataFilter, (),
                  transient=True)
    
    def _actors_default(self):
        pipeline = self.pipeline
        
        lut = tvtk.LookupTable()
        lut.range=(0.0,1.0)
        lut.ramp="linear"
        lut.alpha_range=(1.0,0.0)
        lut.hue_range=(0.65,0.65)
        lut.value_range=(0.4,0.4)
        lut.build()
        lut.below_range_color=(0.2,0.0,0.2,1.0)
        lut.above_range_color=(0.2,0.0,0.2,1.0)
        lut.use_below_range_color=True
        lut.use_above_range_color=True
        
        map = tvtk.PolyDataMapper(input_connection=pipeline.output_port,
                                  lookup_table=lut)
        map.scalar_range = (0.0,1.0)
        #map.scalar_visibility = True
        act = tvtk.Actor(mapper=map)
        #act.property = self.vtkproperty
        actors = tvtk.ActorCollection()
        actors.append(act)
        return actors


class CircularAperture(BaseAperture):
    abstract=False
    """
    This circular aperture is characterised by 3 diameters::
    
      outer_diameter : rays passing outside this diameter do not intersect
      inner_diameter : rays passing within the outer diameter but outside of the inner
                       diameter are blocked (creating no child rays).
                       Rays passing inside the inner_diameter intersect and generate
                       a child ray with amplitudes determined by the position 
                       relative to the aperture origin
      hole_diameter  : Rays passing inside the inner_diameter are attenuated if they 
                       pass outside the hole diameter and left unattenuated
                       if they are inside the hole. The boundary region is generates
                       a sigmoidal profile of width given by the edge_width parameter.
    """
    outer_diameter = Float(25.0)
    inner_diameter = Float(15.0)
    hole_diameter = Float(10.0)
    
    diameter = Property()
    offset = Float(0.0)
    
    name = Str("Circular aperture")
    
    vtkproperty = tvtk.Property(opacity = 0.7,
                             color = (0.0,0.2,0.8))
    
    vtk_disk = Instance(tvtk.DiskSource, (),
                            dict(circumferential_resolution=64,
                                 inner_radius=0), transient=True)
    vtk_disk2 = Instance(tvtk.DiskSource, (),
                            dict(circumferential_resolution=64,
                                 radial_resolution=50,
                                 inner_radius=0), transient=True)
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('outer_diameter', editor=NumEditor),
                       Item('inner_diameter', editor=NumEditor),
                       Item('hole_diameter', editor=NumEditor),
                       Item('edge_width', editor=NumEditor),
                       Item('invert'),
                        ),
                   )
    
    def _get_diameter(self):
        return self.outer_diameter
    
    def _material_default(self):
        m = self.make_material()
        return m
    
    def _faces_default(self):
        fl = FaceList(owner=self)
        fl.faces = self.make_faces()
        return fl
    
    def make_material(self):
        m = CircularApertureMaterial(origin=self.centre,
                                    radius = self.hole_diameter/2,
                                     outer_radius = self.inner_diameter/2,
                                    edge_width = self.edge_width/2,
                                    invert=self.invert)
        return m
    
    def make_faces(self):
        fl = [CircularFace(owner=self, diameter=self.outer_diameter,
                                material = self.material)]
        return fl
    
    @on_trait_change("inner_diameter, hole_diameter, edge_width, centre, invert")
    def on_material_params_changed(self):
        self.material = self.make_material()
        self.faces.faces = self.make_faces()
        self.vtk_disk.inner_radius = self.inner_diameter/2
        self.config_pipeline()
        self.update = True
        
    @on_trait_change("outer_diameter")
    def on_face_params_changed(self):
        self.faces.faces = self.make_faces()
        self.vtk_disk.outer_radius = self.outer_diameter/2
        self.config_pipeline()
        self.update = True
        
    def config_pipeline(self):
        disk = self.vtk_disk
        disk.inner_radius = self.inner_diameter/2
        disk.outer_radius = self.outer_diameter/2
        disk2 = self.vtk_disk2
        disk2.inner_radius = 0
        disk2.outer_radius = self.inner_diameter/2
        self.sc.modified()
        
    def _pipeline_default(self):
        disk1 = self.vtk_disk
        disk2 = self.vtk_disk2
        
        append = tvtk.AppendPolyData()
        append.add_input_connection(disk1.output_port)
        append.add_input_connection(disk2.output_port)
        
        sc = self.sc
        sc.input_connection = append.output_port

        def update( *args ):
            in_data = sc.get_input_data_object(0,0)
            points = in_data.points.to_array()
            centre = numpy.array(self.centre)
            r = numpy.sqrt(((points)**2).sum(axis=-1))
            width = self.edge_width
            r0 = self.hole_diameter
            alpha = 0.5 - 0.5*(erf(2*(r-r0)/width))
            alpha[r>=(0.99*self.inner_diameter/2)] = -0.01
            out = sc.get_output_data_object(0)
            if self.invert:
                alpha = 1 - alpha
            out.point_data.scalars=alpha
            
        sc.set_execute_method(update)
        
        trans = tvtk.TransformFilter(input_connection=sc.output_port, 
                                       transform=self.transform)
        self.config_pipeline()
        return trans
    
    
        
        
class RectangularAperture(BaseAperture):
    abstract=False
    name = Str("Rectangular aperture")
    
    hole_width = Float(2.0)
    hole_height = Float(5.0)
    outer_width = Float(20.0)
    outer_height = Float(15.0)
    inner_width = Float(10.0)
    inner_height = Float(7.0)
    
    offset = Constant(0.0)
    length = Property()
    width = Property()
    
    vtkplane = Instance(tvtk.PlaneSource, (), {'x_resolution':500, 'y_resolution':500})
    vtkbox = Instance(tvtk.Box, ())
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('hole_width', editor=NumEditor),
                       Item('hole_height', editor=NumEditor),
                       Item('inner_width', editor=NumEditor),
                       Item('inner_height', editor=NumEditor),
                       Item('outer_width', editor=NumEditor),
                       Item('outer_height', editor=NumEditor),
                       Item('edge_width', editor=NumEditor),
                       Item('invert'),
                        ),
                   )
    
    def _get_length(self):
        return self.outer_height
    
    def _get_width(self):
        return self.outer_width
    
    def _material_default(self):
        m = self.make_material()
        return m
    
    def _faces_default(self):
        fl = FaceList(owner=self)
        fl.faces = self.make_faces()
        return fl
    
    def make_material(self):
        m = RectangularApertureMaterial(origin=self.centre,
                                    width = self.hole_width,
                                    height = self.hole_height,
                                    outer_width = self.inner_width,
                                    outer_height = self.inner_height,
                                    edge_width = self.edge_width/2,
                                    invert=self.invert)
        return m
    
    def make_faces(self):
        fl = [RectangularFace(owner=self, length=self.outer_width,
                              width=self.outer_height, offset=0.0,
                              z_plane=0.0,
                                material = self.material)]
        return fl
    
    @on_trait_change("inner_width, inner_height, hole_width, hole_height, edge_width, centre, invert")
    def on_material_params_changed(self):
        self.material = self.make_material()
        self.faces.faces = self.make_faces()
        
        self.config_pipeline()
        self.update = True
        
    @on_trait_change("outer_diameter")
    def on_face_params_changed(self):
        self.faces.faces = self.make_faces()
        self.vtk_disk.outer_radius = self.outer_diameter/2
        self.config_pipeline()
        self.update = True
        
    def config_pipeline(self):
        plane = self.vtkplane
        w = self.outer_width/2.
        h = self.outer_height/2.
        plane.origin = (-w, -h, 0.0)
        plane.point1 = (-w, h, 0.0)
        plane.point2 = (w, -h, 0.0)
        
        box = self.vtkbox
        w = self.inner_width/2.
        h = self.inner_height/2.
        box.set_x_min(-w,-h,-1.0)
        box.set_x_max(w,h,1.0)
        
        self.sc.modified()
    
    def _pipeline_default(self):
        plane = self.vtkplane
        
        box = self.vtkbox
        
        clip = tvtk.ClipPolyData(input_connection=plane.output_port)
        clip.clip_function=box
        
        sc = self.sc
        sc.input_connection = plane.output_port

        def update( *args ):
            in_data = sc.get_input_data_object(0,0)
            points = in_data.points.to_array()
            centre = numpy.array(self.centre)
            
            width = self.edge_width
            x0 = self.hole_width/2
            y0 = self.hole_height/2
            x = points[:,0]
            y = points[:,1]
            
            alpha = 0.5 - 0.5*(erf((x-x0)/width))
            alpha *= 0.5 - 0.5*(erf(-(x+x0)/width))
            alpha *= 0.5 - 0.5*(erf((y-y0)/width))
            alpha *= 0.5 - 0.5*(erf(-(y+y0)/width))
            
            x1 = self.inner_width/2
            y1 = self.inner_height/2
            
            alpha[numpy.abs(x)>=x1] = -0.01
            alpha[numpy.abs(y)>=y1] = -0.01
            
            out = sc.get_output_data_object(0)
            if self.invert:
                alpha = 1 - alpha
            out.point_data.scalars=alpha
            
        sc.set_execute_method(update)
        
        trans = tvtk.TransformFilter(input_connection=sc.output_port, 
                                       transform=self.transform)
        self.config_pipeline()
        return trans


'''
Created on 26 Nov 2018

@author: bryan
'''
from traits.api import Float, Int, on_trait_change

from traitsui.api import View, Item, VGroup

from raypier.core.cmaterials import DiffractionGratingMaterial

from raypier.bases import Traceable, normaliseVector, NumEditor,\
     ComplexEditor, VectorEditor, Optic
from raypier.mirrors import RectMirror


class RectangularGrating(RectMirror):
    """
    A reflective diffraction-grating.
    """
    
    #: Object name is display in model-tree
    name = "Rectangular Diffraction Grating"
    
    #: Order of diffraction (int).
    order = Int(1)
    
    #: Lines per mm for the grating
    lines_per_mm = Float(600.0)
    
    #: The diffraction efficiency. A float in range 0.0 -> 1.0 (no loss).
    efficiency = Float(1.0)
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('length', editor=NumEditor),
                       Item('width', editor=NumEditor),
                       Item('thickness', editor=NumEditor),
                       Item('offset', editor=NumEditor),
                       Item('order'),
                       Item('lines_per_mm', editor=NumEditor),
                       Item('efficiency', editor=NumEditor)
                        ),
                   )
    
    @on_trait_change("centre")
    def on_centre_changed(self, vnew):
        self.material.origin = vnew
    
    def _material_default(self):
        return DiffractionGratingMaterial(order=self.order,
                                          origin=self.centre,
                                          lines_per_mm=self.lines_per_mm,
                                          efficiency=self.efficiency)
        
    @on_trait_change("order, lines_per_mm, efficiency")
    def on_parameter_change(self, name, new):
        setattr(self.material, name, new)
        self.update = True
        
    def make_step_shape(self):
        shape, colour = super().make_step_shape()
        return shape, "pink"
        
        


from raypier.splines import Extruded_bezier, b_spline_to_bezier_series
from raypier.sources import RectRaySource
from raypier.tracer import RayTraceModel
from raypier.core.cmaterials import OpaqueMaterial, PECMaterial,DielectricMaterial
from raypier.results import RayPaths
import numpy as np

from scipy.interpolate import splprep, splev

x = np.linspace(1, 30, 100)
# dont use form like this:
#x = np.array([1,2,3,4,5,6,7,8,9,10])
# because these are used as integers and decimals are not used!
y = 30/x 


tck, uout = splprep([x,y], s=.0005, k=3, per=False)
ctrl_pts = b_spline_to_bezier_series(tck)


#give profile as list of x values then y values.  this profile is evaluated by splprep
#the smaller smoothness is, the more accurate the spline is, but the more faces it takes.
#try to make s as small as possible without making too many faces. (20 faces isn't unreasonable)
test = Extruded_bezier(control_points = ctrl_pts, z_height_1 = -30, z_height_2=29, \
		material = DielectricMaterial(), trace_ends= True, trace_top = True, invert_normal=True)
'''
flattest = Extruded_bezier(control_points = np.array([[[0,0],[10,0],[20,0],[30,0]]]), z_height_1 = -30, z_height_2=30, \
		material = DielectricMaterial(), trace_ends= True, trace_top = False, invert_normal=True )
'''

source = RectRaySource(origin=(10,30,0),
                            direction=(.1,-1,0),
                            working_dist = 100.,
                            number=1,
                            length = 0,
            			    width = 0,
            			    randomness = False,
                            scale_factor=0.2)
'''

source = RectRaySource(origin=(10.,10.,0),
                            direction=(.2,-1,0),
                            working_dist = 100.,
                            number=3,
                            length = 12,
            			    width = 12,
            			    randomness = False,
                            scale_factor=0.2)
'''
raypaths = RayPaths()

model = RayTraceModel(sources=[source],
                      optics=[test], results = [raypaths], recursion_limit=10)

#test.invert_normal = True
hits = model.sources[0].traced_rays
res = model.results[0].result
model.configure_traits()





#    Copyright 2009, Teraview Ltd., Bryan Cole
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy
from .utils import normaliseVector, transformPoints
from tvtk.api import tvtk

def transform_pts(centre, direction, points):
    #take a point from the global referance frame and transform it into the 
    #local referance frame of an object.  accepts the x,y,z of the centre and
    #direction of the object's frame.
    
    temp = tvtk.Transform()

    temp.identity()
    temp.translate(centre)

    x,y,z = normaliseVector(direction)
    Theta = numpy.arccos(z)
    theta = 180*Theta/numpy.pi
    phi = 180*numpy.arctan2(x,y)/numpy.pi
    orientation = -phi, -theta
    
    temp.rotate_z(-phi)
    temp.rotate_x(-theta)
    
    inv_t = temp.linear_inverse
    
    result =transformPoints(inv_t, points)
        
    return result

def compute_length(start, end):
    #takes [n,3] vectors and returns [1,n] array of distances
    a = start - end
    d = a**2
    e = sum(d.T)
    distance = numpy.sqrt(e)
    
    mask = distance < .0001
    distance[mask]=numpy.Infinity
    return distance 
    
def interpolate_z(first, threeD, twoD):
        # takes an 3d origin (x,y,z), a 3d point on the line, and intrpolates the third
        # dimesnsion of another point on that line, for which x and y are given 
        # and z is 0.
        # pass numpy.fliplr(P1[:,1:]).copy() to interpolate y instead,
        # and then flip the result back
        
        # len2d1/len2d2 = LenInZ1/LenInZ2
    
    len2d1 = compute_length(first[:,:2], threeD[:,:2])
    len2d2 = compute_length(first[:,:2], twoD[:,:2])
       
    k = len2d2 / len2d1
        
    #Zf = Z1 - k*ChangeinZ         
    z = first[:,2] - (first[:,2]-threeD[:,2])*k
        
    return  z




from traits.api import on_trait_change, Float, Instance,Event, Int,\
        Property, Str, Array, cached_property, List, Bool, observe, Button

from traitsui.api import View, Item, VGroup, Tabbed

from tvtk.api import tvtk

from .bases import Probe, Traceable, NumEditor
from .probes import BaseCapturePlane
from .sources import RayCollection, BaseRaySource

from .core.find_focus import find_ray_focus
from .core.utils import normaliseVector, dotprod
from .core.ctracer import GaussletCollection, RayCollection
from .core.fields import eval_Efield_from_gausslets, eval_Efield_from_rays
from .editors import IntEditor

import numpy
    
import time
import traceback
    
    
class EFieldPlane(Probe):
    #: The name of the object in the model-tree
    name = Str("E-Field Probe")
    
    #: An instance of a :py:class:`raypier.probes.BaseCapturePlane` which provides the
    #: incident rays for which the E-field should be evaluated.
    detector = Instance(BaseCapturePlane)
    
    #: If True, the detector object should be positioned at the centre position for this EFieldPlane
    #: whenever the object moves. 
    align_detector = Bool(True)
    
    #: If no detector-object is given, select the input rays by indexing the source traced_rays list. 
    gen_idx = Int(-1)
    
    #: Width of the EFieldPlane
    width = Float(0.5) #in mm
    
    #: Height of the EField Plane
    height = Float(0.5)
    
    #: The resolution of the EFieldPlane. The same value is used for both x- and y-sides.
    size = Int(30)
            
    #: For RayFields (not Gausslets), be can back-project the indicent rays on to a plane 
    #: and evaluate the Gaussia modes from this "exit pupil" plane. This is ignored for 
    #: the E-field evaluation of Gausslets.
    exit_pupil_offset = Float(10.0) #in mm
    
    #: Controls the overlap of the Gaussian modes at the point of evaluation. It's best to
    #: leave this at unity, otherwise one can get confused between this and the blending factor
    #: offered by the gausslet_source classes.
    blending = Float(1.0)
    
    #: When assigned to (triggered) the EField plane will be repositioned on the geometric focus
    #: of the input rays (calculated as the point of closest approach of the input rays).
    centre_on_focus_btn = Button()
    
    #: The output of the probe. A (size,size,3)-shaped complex array.
    E_field = Array()
    
    #: property - The "|E-field|**2" calculated from the E-field.
    intensity = Property(Array, depends_on="E_field")
    
    #: property - The sum of the intensity array.
    total_power = Property(depends_on="intensity, width, height, size")
    
    #: The mean real part of the refractive index for the material where the probe plane lies
    #: Get get this from the selected/captured rays.
    refractive_index = Float(1.0)
        
    _mtime = Float(0.0)
    _src_list = List()
    _plane_src = Instance(tvtk.PlaneSource, (), 
                    {"x_resolution":1, "y_resolution":1},
                    transient=True)
    
    _attrib = Instance(tvtk.ProgrammableAttributeDataFilter,(), transient=True)
                    
    traits_view = View(Tabbed(
                        VGroup(
                       Traceable.uigroup,
                       Item("total_power", style="readonly"),
                       Item('size', editor=IntEditor),
                       Item('width', editor=NumEditor),
                       Item('height', editor=NumEditor),
                       Item('exit_pupil_offset', editor=NumEditor),
                       Item('blending', editor=NumEditor),
                       Item('gen_idx', editor=IntEditor),
                       Item('centre_on_focus_btn', show_label=False, label="Centre on focus")
                   )))
    
    @observe("centre")
    def on_move(self, evt):
        detector = self.detector
        if detector is not None and self.align_detector:
            detector.centre = evt.new
        self._mtime = 0.0
        self.on_change()
    
    @on_trait_change("orientation, size, width, height, exit_pupil_offset, blending, gen_idx")
    def config_pipeline(self):
        src = self._plane_src
        size = self.size
        side = self.width/2.
        yside = self.height/2
        
        src.origin = (-side,-yside,0)
        src.point1 = (-side,yside,0)
        src.point2 = (side,-yside,0)
        
        src.x_resolution = size
        src.y_resolution = size
        
        self._mtime = 0.0
        self.on_change()
        
    def on_change(self):
        try:
            self.evaluate(None)
        except:
            traceback.print_exc()
        self.update=True
    
    def _actors_default(self):
        source = self._plane_src
        attr = self._attrib
        attr.input_connection = source.output_port
        
        trans_f = tvtk.TransformFilter(input_connection=attr.output_port,
                        transform = self.transform)
        map = tvtk.PolyDataMapper(input_connection=trans_f.output_port)
        
        def execute():
            dataobj = attr.poly_data_output
            data = self.intensity.T
            dataobj.cell_data.scalars = data.ravel()
            map.scalar_range = (data.min(), data.max()) 
        attr.set_execute_method(execute)
        
        act = tvtk.Actor(mapper=map)
        actors = tvtk.ActorCollection()
        actors.append(act)
        return actors
    
    @cached_property
    def _get_intensity(self):        
        E = self.E_field
        return (E.real**2).sum(axis=-1) + (E.imag**2).sum(axis=-1)
    
    def _get_total_power(self):
        power = self.intensity.sum()*(self.width*self.height/(self.size**2))
        power *= self.refractive_index
        return power
        
    def evaluate(self, src_list):
        mtime = self._mtime
        detector = self.detector
        if src_list is None:
            src_list = self._src_list
        else:
            self._src_list = src_list
        
        start = time.monotonic()
            
        if detector is not None:
            detector.evaluate(src_list)
            if detector.captured is None:
                return
            if mtime > detector._mtime:
                return
            ray_list = detector.captured
        else:
            idx = self.gen_idx
            ray_list = [src.traced_rays[idx] for src in src_list if src.traced_rays]
        
        n_list = []
        for ct, rays in enumerate(ray_list):
            wavelengths = rays.wavelengths
            
            size = self.size
            side = self.width/2.
            yside = self.height/2
            px = numpy.linspace(-side,side,size)
            py = numpy.linspace(-yside, yside, size)
            
            centre = numpy.asarray(self.centre)
            axis2 = numpy.cross(self.direction, self.x_axis)
            axis1 = numpy.cross(axis2, self.direction)
            
            points = centre[None,None,:] + px[None,:,None]*axis1 + py[:,None,None]*axis2
            points2 = points.reshape(-1,3)
            
            if isinstance(rays, GaussletCollection):
                n_list.append(rays.base_rays.refractive_index.real)
                E = eval_Efield_from_gausslets(rays, points2,
                                               blending=self.blending)
            else:
                n_list.append(rays.refractive_index.real)
                E = eval_Efield_from_rays(rays, points2, wavelengths, 
                                          blending=self.blending,
                                          exit_pupil_offset=self.exit_pupil_offset,
                                          exit_pupil_centre=self.centre)
            
            if ct==0:
                E_field = E.reshape(self.size, self.size, 3)
            else:
                E_field += E.reshape(self.size, self.size, 3)
                
        if not n_list:
            return
                
        self.refractive_index = numpy.concatenate(n_list).mean()
        self.E_field = E_field
        
        self._attrib.modified()
        end = time.monotonic()
        self._mtime = end
        print(f"Field calculation took: {end-start} s")
        
            
    def intersect_plane(self, rays):
        """
        @param rays: a numpy array of ray_t dtype
        
        intersect rays with plane, returning the indices of the intersecting rays
        """
        #raise NotImplementedError
        pass
    
    @observe("centre_on_focus_btn")
    def do_centre_on_focus(self, evt):
        self.centre_on_focus()
    
    def centre_on_focus(self, idx=-1, offset=(0,0,0)):
        """
        Locate the centre of the field-probe on the point of closest approach 
        for the indicated RayCollection (by default, the last one).
        """
        detector = self.detector
        if detector is None:
            src = self.source
            rays = src.traced_rays[idx]
        else:
            rays = detector.captured[0]
        self.centre = tuple(a+b for a,b in zip(find_ray_focus(rays), offset))
    
    
    